# Copyright 2023-2024 Vlad Krupinskii <mrvladus@yandex.ru>
# SPDX-License-Identifier: MIT

from errands.lib.sync.sync import Sync
from errands.widgets.components import Box
from errands.widgets.details import Details
from gi.repository import Adw, Gtk, GLib, GObject  # type:ignore
from errands.lib.animation import scroll
from errands.lib.data import UserData
from errands.lib.functions import get_children
from errands.lib.logging import Log
from errands.widgets.task import Task
from errands.lib.gsettings import GSettings


class TaskList(Adw.Bin):
    # State
    scrolling: bool = False  # Is window scrolling

    def __init__(self, window, list_uid: str, parent):
        super().__init__()
        self.window = window
        self.list_uid = list_uid
        self.parent = parent
        self._build_ui()
        self.update_ui()

    def _build_ui(self) -> None:
        # Title
        self.title = Adw.WindowTitle(
            title=UserData.run_sql(
                f"SELECT name FROM lists WHERE uid = '{self.list_uid}'",
                fetch=True,
            )[0][0]
        )

        # Toggle sidebar button
        self.left_toggle_sidebar_btn = Gtk.ToggleButton(
            icon_name="errands-sidebar-left-symbolic",
            tooltip_text=_("Toggle Sidebar"),
        )
        toggle_ctrl = Gtk.ShortcutController(scope=1)
        toggle_ctrl.add_shortcut(
            Gtk.Shortcut(
                trigger=Gtk.ShortcutTrigger.parse_string("F9"),
                action=Gtk.ShortcutAction.parse_string("activate"),
            )
        )
        self.left_toggle_sidebar_btn.add_controller(toggle_ctrl)
        self.left_toggle_sidebar_bin = Adw.Bin(child=self.left_toggle_sidebar_btn)
        GSettings.bind("right-sidebar", self.left_toggle_sidebar_btn, "visible", True)

        self.right_toggle_sidebar_btn = Gtk.ToggleButton(
            icon_name="errands-sidebar-right-symbolic",
            tooltip_text=_("Toggle Sidebar"),
        )
        self.right_toggle_sidebar_btn.bind_property(
            "active",
            self.left_toggle_sidebar_btn,
            "active",
            GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL,
        )
        self.right_toggle_sidebar_bin = Adw.Bin(child=self.right_toggle_sidebar_btn)
        GSettings.bind("right-sidebar", self.right_toggle_sidebar_btn, "visible")

        # Delete completed button
        self.delete_completed_btn = Gtk.Button(
            valign="center",
            icon_name="edit-clear-all-symbolic",
            tooltip_text=_("Delete Completed Tasks"),
            sensitive=False,
        )
        self.delete_completed_btn.connect(
            "clicked", self._on_delete_completed_btn_clicked
        )
        # Scroll up btn
        self.scroll_up_btn = Gtk.Button(
            valign="center",
            icon_name="go-up-symbolic",
            tooltip_text=_("Scroll Up"),
            sensitive=False,
        )
        self.scroll_up_btn.connect("clicked", lambda *_: scroll(self.scrl, False))

        # Header Bar
        hb = Adw.HeaderBar(title_widget=self.title)
        hb.pack_start(self.left_toggle_sidebar_bin)
        hb.pack_start(self.delete_completed_btn)
        hb.pack_end(self.right_toggle_sidebar_bin)
        hb.pack_end(self.scroll_up_btn)

        # ---------- BOTTOMBAR ---------- #

        left_toggle_sidebar_btn = Gtk.ToggleButton(
            icon_name="errands-sidebar-left-symbolic",
            tooltip_text=_("Toggle Sidebar"),
        )
        left_toggle_sidebar_btn.bind_property(
            "active",
            self.left_toggle_sidebar_btn,
            "active",
            GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL,
        )
        GSettings.bind("right-sidebar", left_toggle_sidebar_btn, "visible", True)
        right_toggle_sidebar_btn = Gtk.ToggleButton(
            icon_name="errands-sidebar-right-symbolic",
            tooltip_text=_("Toggle Sidebar"),
        )
        right_toggle_sidebar_btn.bind_property(
            "active",
            self.left_toggle_sidebar_btn,
            "active",
            GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL,
        )
        GSettings.bind("right-sidebar", right_toggle_sidebar_btn, "visible")

        delete_completed_btn = Gtk.Button(
            valign="center",
            icon_name="edit-clear-all-symbolic",
            tooltip_text=_("Delete Completed Tasks"),
            sensitive=False,
        )
        delete_completed_btn.bind_property(
            "sensitive",
            self.delete_completed_btn,
            "sensitive",
            GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL,
        )
        delete_completed_btn.connect("clicked", self._on_delete_completed_btn_clicked)

        scroll_up_btn = Gtk.Button(
            valign="center",
            icon_name="go-up-symbolic",
            tooltip_text=_("Scroll Up"),
            sensitive=False,
        )
        scroll_up_btn.bind_property(
            "sensitive",
            self.scroll_up_btn,
            "sensitive",
            GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL,
        )
        scroll_up_btn.connect("clicked", lambda *_: scroll(self.scrl, False))

        # ---------- TASKS LIST ---------- #

        # Srolled window
        adj = Gtk.Adjustment()
        adj.connect(
            "value-changed",
            lambda *_: self.scroll_up_btn.set_sensitive(adj.get_value() > 0),
        )
        self.scrl = Gtk.ScrolledWindow(
            propagate_natural_height=True,
            propagate_natural_width=True,
            vadjustment=adj,
        )
        self.dnd_ctrl = Gtk.DropControllerMotion()
        self.dnd_ctrl.connect("motion", self._on_dnd_scroll)
        self.scrl.add_controller(self.dnd_ctrl)

        # Tasks list
        self.tasks_list = Gtk.Box(orientation="vertical", hexpand=True, vexpand=True)
        self.tasks_list.add_css_class("tasks-list")
        self.scrl.set_child(
            Adw.Clamp(
                maximum_size=1000,
                tightening_threshold=300,
                child=self.tasks_list,
            )
        )

        # Content box
        content_box = Box(
            children=[TaskListEntry(self), self.scrl],
            orientation="vertical",
            vexpand=True,
        )
        content_box_click_ctrl = Gtk.GestureClick()
        content_box_click_ctrl.connect("released", self._on_empty_area_clicked)
        content_box.add_controller(content_box_click_ctrl)

        # Split View
        self.split_view = Adw.OverlaySplitView(
            min_sidebar_width=360,
            max_sidebar_width=400,
            sidebar_width_fraction=0.40,
            sidebar_position=int(GSettings.get("right-sidebar")),
        )
        self.split_view.bind_property(
            "show-sidebar",
            self.left_toggle_sidebar_btn,
            "active",
            GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL,
        )
        GSettings.bind("sidebar-open", self.split_view, "show-sidebar")

        # Details
        self.details = Details(self)
        self.split_view.set_sidebar(self.details)

        # Tasks list toolbar view
        self.tasks_toolbar_view = Adw.ToolbarView(
            content=content_box,
            reveal_bottom_bars=False,
        )
        self.tasks_toolbar_view.add_top_bar(hb)
        self.tasks_toolbar_view.add_bottom_bar(
            Box(
                children=[
                    left_toggle_sidebar_btn,
                    delete_completed_btn,
                    Gtk.Separator(hexpand=True, css_classes=["spacer"]),
                    scroll_up_btn,
                    right_toggle_sidebar_btn,
                ],
                css_classes=["toolbar"],
            )
        )
        self.split_view.set_content(self.tasks_toolbar_view)

        # Breakpoints
        tasks_brb = Adw.BreakpointBin(
            width_request=360, height_request=360, child=self.split_view
        )
        bp = Adw.Breakpoint.new(Adw.breakpoint_condition_parse("max-width: 720px"))
        bp.add_setter(self.split_view, "collapsed", True)
        tasks_brb.add_breakpoint(bp)
        bp1 = Adw.Breakpoint.new(Adw.breakpoint_condition_parse("max-width: 370px"))
        bp1.add_setter(self.split_view, "collapsed", True)
        bp1.add_setter(self.left_toggle_sidebar_bin, "visible", False)
        bp1.add_setter(self.right_toggle_sidebar_bin, "visible", False)
        bp1.add_setter(self.delete_completed_btn, "visible", False)
        bp1.add_setter(self.scroll_up_btn, "visible", False)
        bp1.add_setter(self.tasks_toolbar_view, "reveal-bottom-bars", True)
        tasks_brb.add_breakpoint(bp1)

        self.set_child(tasks_brb)

    def add_task(self, uid: str) -> None:
        new_task = Task(uid, self.list_uid, self.window, self, self, False)
        self.tasks_list.append(new_task)
        new_task.toggle_visibility(not new_task.get_prop("trash"))

    def get_all_tasks(self) -> list[Task]:
        """
        Get list of all tasks widgets including sub-tasks
        """

        tasks: list[Task] = []

        for task in self.get_toplevel_tasks():
            tasks.append(task)
            tasks.extend(task.tasks_list.get_all_sub_tasks())

        return tasks

    def get_toplevel_tasks(self) -> list[Task]:
        return get_children(self.tasks_list)

    def update_status(self) -> None:
        """
        Update status bar on the top
        """

        tasks: list[Task] = self.get_toplevel_tasks()
        n_total: int = len([t for t in tasks if t.get_reveal_child()])
        n_completed: int = len(
            [
                t
                for t in tasks
                if t.task_row.complete_btn.get_active() and t.get_reveal_child()
            ]
        )

        self.title.set_subtitle(
            _("Completed:") + f" {n_completed} / {n_total}" if n_total > 0 else ""
        )
        self.delete_completed_btn.set_sensitive(
            len(
                [
                    t
                    for t in self.get_all_tasks()
                    if t.task_row.complete_btn.get_active() and t.get_reveal_child()
                ]
            )
            > 0
        )

    def update_ui(self) -> None:
        Log.debug(f"Task list {self.list_uid}: Update UI")

        # Rename list
        self.title.set_title(
            UserData.run_sql(
                f"""SELECT name FROM lists 
                WHERE uid = '{self.list_uid}'""",
                fetch=True,
            )[0][0]
        )

        # Remove deleted tasks
        all_tasks: list[Task] = self.get_all_tasks()
        ids: list[str] = UserData.get_tasks_uids(self.list_uid)
        for task in all_tasks:
            if task.uid not in ids:
                task.purge()
                all_tasks.remove(task)

        # Create new tasks
        for task_dict in UserData.get_tasks_as_dicts(self.list_uid):
            if task_dict["uid"] not in [task.uid for task in self.get_all_tasks()]:
                # Add toplevel task and its sub-tasks
                if task_dict["parent"] == "":
                    self.add_task(task_dict["uid"])
                # Add sub-task and its sub-tasks
                else:
                    for t in self.get_all_tasks():
                        if t.uid == task_dict["parent"]:
                            t.tasks_list.add_sub_task(task_dict["uid"])

        # Update details
        if (
            self.details.parent
            and self.details.parent.uid in UserData.get_tasks_uids(self.list_uid)
            and not self.details.parent.get_prop("trash")
        ):
            self.details.update_info(self.details.parent)
        else:
            self.details.update_info(None)

        for task in self.get_toplevel_tasks():
            task.update_ui()

        self.update_status()

    def _on_delete_completed_btn_clicked(self, _) -> None:
        """
        Hide completed tasks and move them to trash
        """
        Log.info("Delete completed tasks")
        for task in self.get_all_tasks():
            if not task.get_prop("trash") and task.get_prop("completed"):
                task.delete()
        self.update_ui()

    def _on_dnd_scroll(self, _motion, _x, y: float) -> bool:
        """
        Autoscroll while dragging task
        """

        def _auto_scroll(scroll_up: bool) -> bool:
            """Scroll while drag is near the edge"""
            if not self.scrolling or not self.dnd_ctrl.contains_pointer():
                return False
            adj: Gtk.Adjustment = self.scrl.get_vadjustment()
            adj.set_value(adj.get_value() - (2 if scroll_up else -2))
            return True

        MARGIN: int = 50
        if y < MARGIN:
            self.scrolling = True
            GLib.timeout_add(100, _auto_scroll, True)
        elif y > self.scrl.get_allocation().height - MARGIN:
            self.scrolling = True
            GLib.timeout_add(100, _auto_scroll, False)
        else:
            self.scrolling = False

    def _on_empty_area_clicked(self, _gesture, _n, x: float, y: float) -> None:
        """Close Details panel on click on empty space"""

        height = 0
        for task in self.get_toplevel_tasks():
            if task.get_child_revealed():
                height = (
                    task.compute_bounds(self.tasks_toolbar_view.get_content())
                    .out_bounds.get_bottom_right()
                    .y
                )
        left_area_end: int = self.tasks_list.get_allocation().x
        right_area_start: int = left_area_end + self.tasks_list.get_width()

        on_sides: bool = x < left_area_end or x > right_area_start
        on_bottom: bool = y > height
        if on_sides or on_bottom:
            self.split_view.set_show_sidebar(False)


class TaskListEntry(Adw.Bin):
    def __init__(self, task_list: TaskList) -> None:
        super().__init__()
        self.task_list = task_list
        self._build_ui()

    def _build_ui(self):
        # Entry
        entry = Adw.EntryRow(
            activatable=False,
            height_request=60,
            title=_("Add new Task"),
        )
        entry.connect("entry-activated", self._on_task_added)
        # Box
        box = Gtk.ListBox(
            css_classes=["boxed-list"],
            selection_mode=Gtk.SelectionMode.NONE,
            margin_top=12,
            margin_bottom=12,
            margin_start=12,
            margin_end=12,
        )
        box.append(entry)
        # Clamp
        self.set_child(
            Adw.Clamp(
                maximum_size=1000,
                tightening_threshold=300,
                child=box,
            )
        )

    def _on_task_added(self, entry: Adw.EntryRow):
        text: str = entry.props.text
        if text.strip(" \n\t") == "":
            return
        self.task_list.add_task(
            UserData.add_task(list_uid=self.task_list.list_uid, text=text)
        )
        entry.props.text = ""
        scroll(self.task_list.scrl, True)
        Sync.sync()



# Copyright 2023 Vlad Krupinskii <mrvladus@yandex.ru>
# SPDX-License-Identifier: MIT

from __future__ import annotations
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from errands.widgets.task_list import TaskList
    from errands.widgets.window import Window

from errands.widgets.components import Box, Button
from gi.repository import Gtk, Adw, Gdk, GObject  # type:ignore
from errands.lib.sync.sync import Sync
from errands.lib.logging import Log
from errands.lib.data import UserData
from errands.lib.markup import Markup
from errands.lib.functions import get_children
from errands.lib.gsettings import GSettings


class TaskTopDropArea(Gtk.Revealer):
    def __init__(self, task: Task):
        super().__init__()
        self.task: Task = task
        self._build_ui()

    def _build_ui(self):
        self.set_transition_type(5)

        # Drop Image
        top_drop_img: Gtk.Image = Gtk.Image(
            icon_name="list-add-symbolic",
            hexpand=True,
            css_classes=["dim-label", "task-drop-area"],
        )
        top_drop_img_target: Gtk.DropTarget = Gtk.DropTarget.new(
            actions=Gdk.DragAction.MOVE, type=Task
        )
        top_drop_img_target.connect("drop", self._on_drop)
        top_drop_img.add_controller(top_drop_img_target)
        self.set_child(top_drop_img)

        # Drop controller
        drop_ctrl: Gtk.DropControllerMotion = Gtk.DropControllerMotion.new()
        drop_ctrl.bind_property(
            "contains-pointer", self, "reveal-child", GObject.BindingFlags.SYNC_CREATE
        )
        self.task.add_controller(drop_ctrl)

    def _on_drop(self, _drop, task: Task, _x, _y) -> bool:
        """
        When task is dropped on "+" area on top of task
        """

        UserData.move_task_to_list(
            task.uid,
            task.list_uid,
            self.task.list_uid,
            self.task.parent.uid if isinstance(self.task.parent, Task) else "",
            False,
        )
        UserData.move_task_before(self.task.list_uid, task.uid, self.task.uid)
        # If task has the same parent
        if task.parent == self.task.parent:
            # Move widget
            self.task.parent.tasks_list.reorder_child_after(task, self.task)
            self.task.parent.tasks_list.reorder_child_after(self.task, task)
            return True
        # Change parent if different parents
        UserData.update_props(
            self.task.list_uid,
            task.uid,
            ["parent", "synced"],
            [self.task.parent.uid if isinstance(self.task.parent, Task) else "", False],
        )
        # Add new task widget
        new_task: Task = Task(
            task.uid,
            self.task.list_uid,
            self.task.window,
            self.task.task_list,
            self.task.parent,
            self.task.get_prop("parent") != None,
        )
        self.task.parent.tasks_list.append(new_task)
        self.task.parent.tasks_list.reorder_child_after(new_task, self.task)
        self.task.parent.tasks_list.reorder_child_after(self.task, new_task)
        new_task.toggle_visibility(True)
        # Toggle completion
        if not task.task_row.complete_btn.get_active():
            self.task.update_props(["completed", "synced"], [False, False])
            self.task.just_added = True
            self.task.task_row.complete_btn.set_active(False)
            self.task.just_added = False
            for parent in self.task.get_parents_tree():
                if parent.task_row.complete_btn.get_active():
                    parent.update_props(["completed", "synced"], [False, False])
                    parent.just_added = True
                    parent.task_row.complete_btn.set_active(False)
                    parent.just_added = False
        # Update status
        task.purge()
        self.task.task_list.update_ui()
        # Sync
        Sync.sync()

        return True


class TaskTitleRow(Gtk.Overlay):
    def __init__(self, task: Task):
        super().__init__()
        self.task: Task = task
        self._build_ui()

    def add_rm_crossline(self, add: bool) -> None:
        if add:
            self.task_row.add_css_class("task-completed")
        else:
            self.task_row.remove_css_class("task-completed")

    def _build_ui(self):
        # Task Title Row
        self.task_row = Adw.ActionRow(
            title=Markup.find_url(Markup.escape(self.task.get_prop("text"))),
            css_classes=["rounded-corners", "transparent"],
            height_request=60,
            accessible_role=Gtk.AccessibleRole.ROW,
            cursor=Gdk.Cursor.new_from_name("pointer"),
            use_markup=True,
        )
        self.add_rm_crossline(self.task.get_prop("completed"))

        # Drag controller
        task_row_drag_source: Gtk.DragSource = Gtk.DragSource.new()
        task_row_drag_source.set_actions(Gdk.DragAction.MOVE)
        task_row_drag_source.connect("prepare", self._on_drag_prepare)
        task_row_drag_source.connect("drag-begin", self._on_drag_begin)
        task_row_drag_source.connect("drag-cancel", self._on_drag_end)
        task_row_drag_source.connect("drag-end", self._on_drag_end)
        self.task_row.add_controller(task_row_drag_source)

        # Drop controller
        task_row_drop_target = Gtk.DropTarget.new(
            actions=Gdk.DragAction.MOVE, type=Task
        )
        task_row_drop_target.connect("drop", self._on_drop)
        self.task_row.add_controller(task_row_drop_target)

        # Click controller
        task_row_click_ctrl = Gtk.GestureClick.new()
        task_row_click_ctrl.connect("released", self._on_row_clicked)
        self.task_row.add_controller(task_row_click_ctrl)

        # Prefix
        self.complete_btn = TaskCompleteButton(self.task, self.task_row)
        self.task_row.add_prefix(self.complete_btn)

        # Suffix
        self.details_btn = TaskDetailsButton(self.task)
        self.expand_btn = TaskExpandButton(self.task)
        self.task_row.add_suffix(Box(children=[self.expand_btn, self.details_btn]))

        self.expand_indicator = Gtk.Image(
            icon_name="errands-up-symbolic",
            css_classes=["expand-indicator"],
            halign=Gtk.Align.END,
            margin_end=50,
        )
        expand_indicator_rev = Gtk.Revealer(
            child=self.expand_indicator,
            transition_type=1,
            reveal_child=False,
            can_target=False,
        )
        GSettings.bind("primary-action-show-sub-tasks", expand_indicator_rev, "visible")
        self.add_overlay(expand_indicator_rev)

        # Expand indicator hover controller
        hover_ctrl = Gtk.EventControllerMotion()
        hover_ctrl.bind_property(
            "contains-pointer",
            expand_indicator_rev,
            "reveal-child",
            GObject.BindingFlags.SYNC_CREATE,
        )
        self.task_row.add_controller(hover_ctrl)

        box = Gtk.ListBox(
            selection_mode=Gtk.SelectionMode.NONE,
            css_classes=["rounded-corners", "transparent"],
            accessible_role=Gtk.AccessibleRole.PRESENTATION,
        )
        box.append(self.task_row)

        self.set_child(box)

    def _on_row_clicked(self, *args) -> None:
        # Show sub-tasks if this is primary action
        if GSettings.get("primary-action-show-sub-tasks"):
            self.task.expand(not self.task.sub_tasks_revealer.get_child_revealed())
        else:
            self.task.task_row.details_btn.do_clicked()

    def update_ui(self):
        # Update widget title crossline and completed toggle
        completed: bool = self.task.get_prop("completed")
        self.add_rm_crossline(completed)
        if self.complete_btn.get_active() != completed:
            self.task.just_added = True
            self.complete_btn.set_active(completed)
            self.task.just_added = False

        # Update subtitle
        total, completed = self.task.get_status()
        self.task_row.set_subtitle(
            _("Completed:") + f" {completed} / {total}" if total > 0 else ""
        )

    # --- DND --- #

    def _on_drag_prepare(self, *_) -> Gdk.ContentProvider:
        # Bug workaround when task is not sensitive after short dnd
        for task in self.task.task_list.get_all_tasks():
            task.set_sensitive(True)
        self.task.set_sensitive(False)
        value: GObject.Value = GObject.Value(Task)
        value.set_object(self.task)
        return Gdk.ContentProvider.new_for_value(value)

    def _on_drag_begin(self, _, drag) -> bool:
        text: str = self.task.get_prop("text")
        icon: Gtk.DragIcon = Gtk.DragIcon.get_for_drag(drag)
        icon.set_child(Gtk.Button(label=text if len(text) < 20 else f"{text[0:20]}..."))

    def _on_drag_end(self, *_) -> bool:
        self.task.set_sensitive(True)
        # KDE dnd bug workaround for issue #111
        for task in self.task.task_list.get_all_tasks():
            task.top_drop_area.set_reveal_child(False)
            task.set_sensitive(True)

    def _on_drop(self, _drop, task: Task, _x, _y) -> None:
        """
        When task is dropped on task and becomes sub-task
        """

        if task == self.task or task.parent == self.task:
            return

        # Change parent
        UserData.move_task_to_list(
            task.uid,
            task.list_uid,
            self.task.list_uid,
            self.task.get_prop("uid"),
            False,
        )
        # Add new sub-task
        self.task.tasks_list.add_sub_task(task.uid)
        # Toggle completion
        if not task.task_row.complete_btn.get_active():
            self.task.update_props(["completed", "synced"], [False, False])
            self.task.just_added = True
            self.task.task_row.complete_btn.set_active(False)
            self.task.just_added = False
            for parent in self.task.get_parents_tree():
                if parent.task_row.complete_btn.get_active():
                    parent.update_props(["completed", "synced"], [False, False])
                    parent.just_added = True
                    parent.task_row.complete_btn.set_active(False)
                    parent.just_added = False
        if not self.task.get_prop("expanded"):
            self.task.expand(True)
        # Remove old task
        task.purge()
        self.task.task_list.update_ui()
        # Sync
        Sync.sync()

        return True


class TaskCompleteButton(Gtk.Box):
    def __init__(self, task: Task, parent: Adw.ActionRow) -> None:
        super().__init__()
        self.task: Task = task
        self.parent: Adw.ActionRow = parent
        self._build_ui()

    def get_active(self) -> bool:
        return self.big_btn.get_active()

    def set_active(self, active: bool) -> None:
        self.big_btn.set_active(active)

    def _build_ui(self) -> None:
        # Big button
        self.big_btn = Gtk.CheckButton(
            valign=Gtk.Align.CENTER,
            tooltip_text=_("Mark as Completed"),
            active=self.task.get_prop("completed"),
            css_classes=["selection-mode"],
        )
        GSettings.bind("task-big-toggle", self.big_btn, "visible")
        self.big_btn.connect("toggled", self._on_toggle)
        self.append(self.big_btn)

        # Small button
        small_btn = Gtk.CheckButton(
            valign=Gtk.Align.CENTER,
            tooltip_text=_("Mark as Completed"),
            active=self.task.get_prop("completed"),
        )
        small_btn.bind_property(
            "active",
            self.big_btn,
            "active",
            GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL,
        )
        GSettings.bind("task-big-toggle", small_btn, "visible", True)
        self.append(small_btn)

    def _on_toggle(self, btn: Gtk.CheckButton) -> None:
        Log.debug(f"Task '{self.task.uid}': Set completed to '{self.get_active()}'")

        # If task is just added set crossline and return to avoid sync loop
        self.task.task_row.add_rm_crossline(self.get_active())
        if self.task.just_added:
            return

        self.task.update_props(["completed", "synced"], [self.get_active(), False])
        sub_tasks: list[Task] = self.task.tasks_list.get_all_sub_tasks()
        parents: list[Task] = self.task.get_parents_tree()

        if self.get_active():
            # Complete all sub-tasks
            for sub in sub_tasks:
                if not sub.task_row.complete_btn.get_active():
                    sub.just_added = True
                    sub.task_row.complete_btn.set_active(True)
                    sub.just_added = False
                    sub.update_props(["completed", "synced"], [True, False])
        else:
            # Uncomplete parent if sub-task is uncompleted
            for parent in parents:
                if parent.task_row.complete_btn.get_active():
                    parent.just_added = True
                    parent.task_row.complete_btn.set_active(False)
                    parent.just_added = False
                    parent.update_props(["completed", "synced"], [False, False])

        # Calculate completion percent for parents
        # for parent in parents:
        #     total, completed = parent.get_status()
        #     pc: int = (
        #         completed / total * 100
        #         if total > 0
        #         else parent.get_prop("percent_complete")
        #     )
        #     if parent.get_prop("percent_complete") != pc:
        #         parent.update_props(["percent_complete", "synced"], [pc, False])
        #     parent.update_ui()

        # Calculate completion percent for self
        # total, completed = self.task.get_status()
        # pc: int = (
        #     completed / total * 100
        #     if total > 0
        #     else (100 if self.task.get_prop("completed") else 0)
        # )
        # if self.task.get_prop("percent_complete") != pc:
        #     self.task.update_props(["percent_complete", "synced"], [pc, False])

        # # Calculate completion percent for sub-tasks
        # for sub in sub_tasks:
        #     total, completed = sub.get_status()
        #     pc: int = (
        #         completed / total * 100
        #         if total > 0
        #         else (100 if self.task.get_prop("completed") else 0)
        #     )
        #     if sub.get_prop("percent_complete") != pc:
        #         sub.update_props(["percent_complete", "synced"], [pc, False])

        self.task.task_list.update_ui()
        Sync.sync()


class TaskExpandButton(Gtk.Button):
    def __init__(self, task: Task):
        super().__init__()
        self.task: Task = task
        self._build_ui()

    def _build_ui(self):
        self.set_icon_name("errands-up-symbolic")
        self.set_valign(Gtk.Align.CENTER)
        self.set_tooltip_text(_("Expand / Fold"))
        self.add_css_class("flat")
        self.add_css_class("circular")
        self.add_css_class("fade")
        self.add_css_class("rotate")
        GSettings.bind("primary-action-show-sub-tasks", self, "visible", True)

    def do_clicked(self):
        self.task.expand(not self.task.sub_tasks_revealer.get_child_revealed())


class TaskDetailsButton(Gtk.Button):
    def __init__(self, task: Task):
        super().__init__()
        self.task: Task = task
        self._build_ui()

    def _build_ui(self):
        self.set_icon_name("errands-info-symbolic")
        self.set_valign(Gtk.Align.CENTER)
        self.set_tooltip_text(_("Details"))
        self.add_css_class("flat")
        self.add_css_class("circular")
        GSettings.bind("primary-action-show-sub-tasks", self, "visible")

    def do_clicked(self):
        # Close details on second click
        if (
            self.task.details.parent == self.task
            and not self.task.details.status.get_visible()
            and self.task.task_list.split_view.get_show_sidebar()
        ):
            self.task.task_list.split_view.set_show_sidebar(False)
            return
        # Update details and show sidebar
        self.task.details.update_info(self.task)
        self.task.task_list.split_view.set_show_sidebar(True)


class TaskInfoBar(Gtk.Box):
    def __init__(self, task: Task):
        super().__init__()
        self.task: Task = task
        self._build_ui()

    def _build_ui(self):
        self.set_orientation(Gtk.Orientation.VERTICAL)

        # Progress bar
        self.progress_bar = Gtk.ProgressBar(
            margin_end=12,
            margin_start=12,
            margin_bottom=6,
        )
        self.progress_bar.add_css_class("osd")
        self.progress_bar.add_css_class("dim-label")
        GSettings.bind("task-show-progressbar", self.progress_bar, "visible")
        self.progress_bar_rev = Gtk.Revealer(
            child=self.progress_bar, transition_duration=100
        )
        self.append(self.progress_bar_rev)

        # Info
        # self.due_date = Button(
        #     icon_name="errands-calendar-symbolic",
        #     label=_("Due"),
        #     css_classes=["task-toolbar-btn"],
        #     cursor=Gdk.Cursor(name="pointer"),
        # )
        # self.priority = Button(
        #     icon_name="errands-priority-symbolic",
        #     label=_("Priority"),
        #     css_classes=["task-toolbar-btn"],
        #     cursor=Gdk.Cursor(name="pointer"),
        # )
        # self.notes = Button(
        #     icon_name="errands-notes-symbolic",
        #     label=_("Notes"),
        #     css_classes=["task-toolbar-btn"],
        #     cursor=Gdk.Cursor(name="pointer"),
        # )
        # self.color = Button(
        #     icon_name="errands-color-symbolic",
        #     label=_("Color"),
        #     css_classes=["task-toolbar-btn"],
        #     cursor=Gdk.Cursor(name="pointer"),
        # )
        # self.tags = Gtk.Box(hexpand=True)
        # self.status_box = Box(
        #     children=[
        #         self.due_date,
        #         self.priority,
        #         self.tags,
        #         self.notes,
        #         self.color,
        #     ],
        #     margin_end=12,
        #     margin_start=12,
        #     margin_bottom=3,
        # )
        # GSettings.bind("task-show-toolbar", self.status_box, "visible")
        # self.append(self.status_box)

    def update_ui(self):
        # end_date = self.task.get_prop("end_date")
        # start_date = self.task.get_prop("start_date")
        # notes = self.task.get_prop("notes")

        # Update percrent complete
        if GSettings.get("task-show-progressbar"):
            total, completed = self.task.get_status()
            pc: int = (
                completed / total * 100
                if total > 0
                else self.task.get_prop("percent_complete")
            )
            if self.task.get_prop("percent_complete") != pc:
                self.task.update_props(["percent_complete", "synced"], [pc, False])

            self.progress_bar.set_fraction(pc / 100)
            self.progress_bar_rev.set_reveal_child(self.task.get_status()[0] > 0)


class TaskSubTasksEntry(Gtk.Entry):
    def __init__(self, task: Task):
        super().__init__()
        self.task: Task = task
        self._build_ui()

    def _build_ui(self):
        self.set_hexpand(True)
        self.set_placeholder_text(_("Add new Sub-Task"))
        self.set_margin_start(12)
        self.set_margin_end(12)
        self.set_margin_bottom(2)

    def do_activate(self) -> None:
        """
        Add new Sub-Task
        """
        text: str = self.get_text()

        # Return if entry is empty
        if text.strip(" \n\t") == "":
            return

        # Add sub-task
        self.task.tasks_list.add_sub_task(
            UserData.add_task(
                list_uid=self.task.list_uid, text=text, parent=self.task.uid
            )
        )

        # Clear entry
        self.set_text("")

        # Update status
        self.task.update_props(["completed"], [False])
        self.task.just_added = True
        self.task.task_row.complete_btn.set_active(False)
        self.task.just_added = False
        self.task.update_ui()

        # Sync
        Sync.sync()


class TaskSubTasks(Gtk.Box):
    def __init__(self, task: Task):
        super().__init__()
        self.task: Task = task
        self._build_ui()
        self._add_sub_tasks()

    def _build_ui(self):
        self.set_orientation(Gtk.Orientation.VERTICAL)
        self.add_css_class("sub-tasks")

    def _add_sub_tasks(self) -> None:
        subs: list[str] = UserData.get_tasks_uids(self.task.list_uid, self.task.uid)
        if len(subs) == 0:
            self.task.just_added = False
            return
        for uid in subs:
            self.add_sub_task(uid)
        # self.task.parent.update_ui()
        self.task.task_list.update_status()
        self.task.just_added = False

    def add_sub_task(self, uid: str) -> Task:
        new_task: Task = Task(
            uid,
            self.task.list_uid,
            self.task.window,
            self.task.task_list,
            self.task,
            True,
        )
        self.append(new_task)
        new_task.toggle_visibility(not new_task.get_prop("trash"))
        return new_task

    def get_sub_tasks(self) -> list[Task]:
        return get_children(self)

    def get_all_sub_tasks(self) -> list[Task]:
        """
        Get list of all tasks widgets including sub-tasks
        """

        tasks: list[Task] = []

        def _append_tasks(sub_tasks: list[Task]) -> None:
            for task in sub_tasks:
                tasks.append(task)
                children: list[Task] = task.tasks_list.get_sub_tasks()
                if len(children) > 0:
                    _append_tasks(children)

        _append_tasks(self.get_sub_tasks())
        return tasks


class Task(Gtk.Revealer):
    just_added: bool = True
    can_sync: bool = True

    def __init__(
        self,
        uid: str,
        list_uid: str,
        window: Window,
        task_list: TaskList,
        parent: TaskList | Task,
        is_sub_task: bool,
    ) -> None:
        super().__init__()
        Log.info(f"Add task: {uid}")

        self.uid = uid
        self.list_uid = list_uid
        self.window = window
        self.task_list = task_list
        self.parent = parent
        self.is_sub_task = is_sub_task
        self.trash = window.trash
        self.details = task_list.details

        self._build_ui()
        # Add to trash if needed
        if self.get_prop("trash"):
            self.trash.trash_add(self)
        # Expand
        self.expand(self.get_prop("expanded"))
        self.update_ui()

    def _build_ui(self) -> None:
        # Top drop area
        self.top_drop_area = TaskTopDropArea(self)

        # Task row
        self.task_row = TaskTitleRow(self)

        # Info bar
        self.info_bar = TaskInfoBar(self)

        # Sub-tasks
        self.tasks_list = TaskSubTasks(self)

        # Sub-tasks revealer
        self.sub_tasks_revealer = Gtk.Revealer(
            child=Box(
                children=[TaskSubTasksEntry(self), self.tasks_list],
                orientation="vertical",
            )
        )

        # Task card
        self.main_box = Box(
            children=[self.task_row, self.info_bar, self.sub_tasks_revealer],
            orientation="vertical",
            hexpand=True,
            css_classes=["fade", "card", f'task-{self.get_prop("color")}'],
        )

        self.set_child(
            Box(
                children=[self.top_drop_area, self.main_box],
                orientation="vertical",
                margin_start=12,
                margin_end=12,
                margin_bottom=6,
                margin_top=6,
            )
        )

    def get_parents_tree(self) -> list[Task]:
        """Get parent tasks chain"""

        parents: list[Task] = []

        def _add(task: Task):
            if isinstance(task.parent, Task):
                parents.append(task.parent)
                _add(task.parent)

        _add(self)

        return parents

    def get_prop(self, prop: str) -> Any:
        res: Any = UserData.get_prop(self.list_uid, self.uid, prop)
        if prop in "deleted completed expanded trash":
            res = bool(res)
        return res

    def get_status(self) -> tuple[int, int]:
        """Get total tasks and completed tasks tuple"""

        total: int = 0
        completed: int = 0
        sub_tasks: list[Task] = self.tasks_list.get_sub_tasks()
        for task in sub_tasks:
            if task.get_reveal_child():
                total += 1
                if task.task_row.complete_btn.get_active():
                    completed += 1

        return total, completed

    def delete(self, *_) -> None:
        """Move task to trash"""

        Log.info(f"Task: Move to trash: '{self.uid}'")

        self.toggle_visibility(False)
        self.update_props(["trash"], [True])
        self.task_row.complete_btn.set_active(True)
        self.trash.trash_add(self)
        for task in self.tasks_list.get_sub_tasks():
            if not task.get_prop("trash"):
                task.delete()
        self.parent.update_ui()

    def expand(self, expanded: bool) -> None:
        self.sub_tasks_revealer.set_reveal_child(expanded)
        self.update_props(["expanded"], [expanded])
        if expanded:
            self.task_row.expand_btn.remove_css_class("rotate")
            self.task_row.expand_indicator.remove_css_class("expand-indicator-expanded")
        else:
            self.task_row.expand_btn.add_css_class("rotate")
            self.task_row.expand_indicator.add_css_class("expand-indicator-expanded")

    def purge(self) -> None:
        """Completely remove widget"""

        self.parent.tasks_list.remove(self)
        self.run_dispose()

    def toggle_visibility(self, on: bool) -> None:
        self.set_reveal_child(on)

    def update_props(self, props: list[str], values: list[Any]) -> None:
        UserData.update_props(self.list_uid, self.uid, props, values)

    def update_ui(self) -> None:
        # Change parent
        parent: str = self.get_prop("parent")
        uid: str = self.uid
        if isinstance(self.parent, Task) and self.parent.uid != parent:
            self.purge()
            if not parent:
                self.task_list.add_task(uid)
                return
            else:
                for task in self.task_list.get_all_tasks():
                    if task.uid == parent:
                        task.add_task(uid)
                        return

        # Change visibility
        self.toggle_visibility(not self.get_prop("trash"))

        self.task_row.update_ui()
        self.info_bar.update_ui()

        for task in self.tasks_list.get_sub_tasks():
            task.update_ui()



import unittest

# Import test modules
import errands.tests.widgets.test_window


def load_tests():
    tests: list[unittest.TestSuite] = []
    loader = unittest.TestLoader()
    tests.append(loader.loadTestsFromModule(errands.tests.widgets.test_window))
    return tests


def run_tests():
    unittest.TextTestRunner().run(unittest.TestSuite(load_tests()))



import datetime
import urllib3
from caldav import Calendar, DAVClient, Principal, Todo
from errands.lib.gsettings import GSettings
from errands.lib.logging import Log
from errands.lib.data import UserData
from gi.repository import Adw, GLib  # type:ignore
from caldav.elements import dav


class SyncProviderCalDAV:
    can_sync: bool = False
    calendars: list[Calendar] = None
    window = None

    def __init__(
        self, window: Adw.ApplicationWindow, testing: bool, name: str = "CalDAV"
    ) -> bool:
        Log.info(f"Sync: Initialize '{name}' sync provider")

        self.name: str = name
        self.window: Adw.ApplicationWindow = window
        self.testing: bool = testing  # Only for connection test

        if not self._check_credentials():
            return

        self._check_url()
        return self._connect()

    def _check_credentials(self) -> bool:
        Log.debug("Sync: Checking credentials")

        self.url: str = GSettings.get("sync-url")
        self.username: str = GSettings.get("sync-username")
        self.password: str = GSettings.get_secret(self.name)

        if self.url == "" or self.username == "" or self.password == "":
            Log.error(f"Sync: Not all {self.name} credentials provided")
            if not self.testing:
                self.window.add_toast(
                    _("Not all sync credentials provided. Please check settings.")
                )
            return False

        return True

    def _check_url(self) -> None:
        Log.debug("Sync: Checking URL")

        self.url = GSettings.get("sync-url")

        Log.debug(f"Sync: URL is set to {self.url}")

    def _connect(self) -> bool:
        Log.debug(f"Sync: Attempting connection")

        urllib3.disable_warnings()

        with DAVClient(
            url=self.url,
            username=self.username,
            password=self.password,
            ssl_verify_cert=False,
        ) as client:
            try:
                self.principal: Principal = client.principal()
                Log.info(f"Sync: Connected to {self.name} server at '{self.url}'")
                self.can_sync = True
                self.calendars = [
                    cal
                    for cal in self.principal.calendars()
                    if "VTODO" in cal.get_supported_components()
                ]
            except Exception as e:
                Log.error(
                    f"Sync: Can't connect to {self.name} server at '{self.url}'. {e}"
                )
                if not self.testing:
                    self.window.add_toast(
                        _("Can't connect to CalDAV server at:")  # pyright:ignore
                        + " "
                        + self.url
                    )

    def _get_tasks(self, calendar: Calendar) -> list[dict]:
        """
        Get todos from calendar and convert them to dict
        """

        try:
            Log.debug(f"Sync: Getting tasks for {calendar.id}")
            todos: list[Todo] = calendar.todos(include_completed=True)
            tasks: list[dict] = []
            for todo in todos:
                data: dict = {
                    "color": str(todo.icalendar_component.get("x-errands-color", "")),
                    "completed": str(todo.icalendar_component.get("status", ""))
                    == "COMPLETED",
                    "notes": str(todo.icalendar_component.get("description", "")),
                    "parent": str(todo.icalendar_component.get("related-to", "")),
                    "percent_complete": int(
                        todo.icalendar_component.get("percent-complete", 0)
                    ),
                    "priority": int(todo.icalendar_component.get("priority", 0)),
                    "text": str(todo.icalendar_component.get("summary", "")),
                    "uid": str(todo.icalendar_component.get("uid", "")),
                }
                # Set tags
                # It is super readable, I know. It works, okay?!
                if (tags := todo.icalendar_component.get("categories", "")) != "":
                    data["tags"] = ",".join(
                        [
                            i.to_ical().decode("utf-8")
                            for i in (tags if isinstance(tags, list) else tags.cats)
                        ]
                    )
                else:
                    data["tags"] = ""
                # Set date
                if todo.icalendar_component.get("due", "") != "":
                    data["end_date"] = (
                        todo.icalendar_component.get("due", "")
                        .to_ical()
                        .decode("utf-8")
                        .strip("Z")
                    )
                else:
                    data["end_date"] = ""

                if todo.icalendar_component.get("dtstart", "") != "":
                    data["start_date"] = (
                        todo.icalendar_component.get("dtstart", "")
                        .to_ical()
                        .decode("utf-8")
                        .strip("Z")
                    )
                else:
                    data["start_date"] = ""
                tasks.append(data)

            # Clear parent of the orphaned tasks
            ids: list[str] = [task["uid"] for task in tasks]
            par_ids: list[str] = [
                task["parent"] for task in tasks if task["parent"] != ""
            ]
            orph_ids: list[str] = [t for t in par_ids if t not in ids]
            for o in orph_ids:
                Log.debug(f"Sync: Delete orphaned task: {o}")
                todo = calendar.todo_by_uid(o)
                todo.icalendar_component["RELATED-TO"] = ""
                todo.save()
            for task in tasks:
                if task["parent"] in orph_ids:
                    task["parent"] = ""

            return [task for task in tasks if task["uid"] not in orph_ids]
        except Exception as e:
            Log.error(f"Sync: Can't get tasks from remote. {e}")
            return []

    def _update_calendars(self) -> bool:
        try:
            self.calendars = [
                cal
                for cal in self.principal.calendars()
                if "VTODO" in cal.get_supported_components()
            ]
            return True
        except:
            Log.error(f"Sync: Can't get caldendars from remote")
            return False

    def sync(self) -> None:
        Log.info(f"Sync: Sync tasks with remote")

        if not self._update_calendars():
            return

        remote_lists_uids = [c.id for c in self.calendars]
        for list in UserData.get_lists_as_dicts():
            for cal in self.calendars:
                # Rename list on remote
                if (
                    cal.id == list["uid"]
                    and cal.name != list["name"]
                    and not list["synced"]
                ):
                    Log.debug(f"Sync: Rename remote list '{list['uid']}'")
                    cal.set_properties([dav.DisplayName(list["name"])])
                    GLib.idle_add(
                        UserData.run_sql,
                        f"UPDATE lists SET synced = 1 WHERE uid = '{cal.id}'",
                    )
                # Rename local list
                elif (
                    cal.id == list["uid"]
                    and cal.name != list["name"]
                    and list["synced"]
                ):
                    Log.debug(f"Sync: Rename local list '{list['uid']}'")
                    GLib.idle_add(
                        UserData.run_sql,
                        f"UPDATE lists SET name = '{cal.name}', synced = 1 WHERE uid = '{cal.id}'",
                    )

            # Delete local list deleted on remote
            if (
                list["uid"] not in remote_lists_uids
                and list["synced"]
                and not list["deleted"]
            ):
                Log.debug(f"Sync: Delete local list deleted on remote '{list['uid']}'")
                GLib.idle_add(
                    UserData.run_sql,
                    f"""DELETE FROM lists WHERE uid = '{list["uid"]}'""",
                )

            # Delete remote list deleted locally
            elif (
                list["uid"] in remote_lists_uids and list["deleted"] and list["synced"]
            ):
                for cal in self.calendars:
                    if cal.id == list["uid"]:
                        Log.debug(f"Sync: Delete list on remote {cal.id}")
                        cal.delete()
                        GLib.idle_add(
                            UserData.run_sql,
                            f"DELETE FROM lists WHERE uid = '{cal.id}'",
                        )
                        break

            # Create new remote list
            elif (
                list["uid"] not in remote_lists_uids
                and not list["synced"]
                and not list["deleted"]
            ):
                Log.debug(f"Sync: Create list on remote {list['uid']}")
                self.principal.make_calendar(
                    cal_id=list["uid"],
                    supported_calendar_component_set=["VTODO"],
                    name=list["name"],
                )
                GLib.idle_add(
                    UserData.run_sql,
                    f"""UPDATE lists SET synced = 1
                    WHERE uid = '{list['uid']}'""",
                )

        if not self._update_calendars():
            return

        user_lists_uids = [i["uid"] for i in UserData.get_lists_as_dicts()]
        remote_lists_uids = [c.id for c in self.calendars]

        for calendar in self.calendars:
            # Get tasks
            local_tasks = UserData.get_tasks_as_dicts(calendar.id)
            local_ids = [t["uid"] for t in local_tasks]
            remote_tasks = self._get_tasks(calendar)
            remote_ids = [task["uid"] for task in remote_tasks]
            deleted_uids = [
                i[0]
                for i in UserData.run_sql(
                    "SELECT uid FROM tasks WHERE deleted = 1", fetch=True
                )
            ]

            # Add new local lists
            if calendar.id not in user_lists_uids:
                Log.debug(f"Sync: Copy list from remote {calendar.id}")
                UserData.add_list(name=calendar.name, uuid=calendar.id, synced=True)

            for task in local_tasks:
                # Update local task that was changed on remote
                if task["uid"] in remote_ids and task["synced"]:
                    for remote_task in remote_tasks:
                        if remote_task["uid"] == task["uid"]:
                            updated: bool = False
                            for key in task.keys():
                                if (
                                    key not in "deleted list_uid synced expanded trash"
                                    and task[key] != remote_task[key]
                                ):
                                    UserData.update_props(
                                        calendar.id,
                                        task["uid"],
                                        [key],
                                        [remote_task[key]],
                                    )
                                    updated = True
                            if updated:
                                Log.debug(
                                    f"Sync: Update local task from remote: {task['uid']}"
                                )
                            break

                # Create new task on remote that was created offline
                if task["uid"] not in remote_ids and not task["synced"]:
                    try:
                        Log.debug(f"Sync: Create new task on remote: {task['uid']}")
                        new_todo = calendar.save_todo(
                            categories=task["tags"] if task["tags"] != "" else None,
                            description=task["notes"],
                            dtstart=(
                                datetime.datetime.fromisoformat(task["start_date"])
                                if task["start_date"]
                                else None
                            ),
                            due=(
                                datetime.datetime.fromisoformat(task["end_date"])
                                if task["end_date"]
                                else None
                            ),
                            priority=task["priority"],
                            percent_complete=task["percent_complete"],
                            related_to=task["parent"],
                            summary=task["text"],
                            uid=task["uid"],
                            x_errands_color=task["color"],
                        )
                        if task["completed"]:
                            new_todo.complete()
                        UserData.update_props(
                            calendar.id, task["uid"], ["synced"], [True]
                        )
                    except Exception as e:
                        Log.error(
                            f"Sync: Can't create new task on remote: {task['uid']}\n{e}"
                        )

                # Update task on remote that was changed locally
                elif task["uid"] in remote_ids and not task["synced"]:
                    try:
                        Log.debug(f"Sync: Update task on remote: {task['uid']}")
                        todo = calendar.todo_by_uid(task["uid"])
                        todo.uncomplete()
                        todo.icalendar_component["summary"] = task["text"]
                        if task["end_date"]:
                            todo.icalendar_component["due"] = task["end_date"]
                        else:
                            todo.icalendar_component.pop("DUE", None)
                        if task["start_date"]:
                            todo.icalendar_component["dtstart"] = task["start_date"]
                        else:
                            todo.icalendar_component.pop("DTSTART", None)
                        todo.icalendar_component["percent-complete"] = task[
                            "percent_complete"
                        ]
                        todo.icalendar_component["description"] = task["notes"]
                        todo.icalendar_component["priority"] = task["priority"]
                        if task["tags"] != "":
                            todo.icalendar_component["categories"] = task["tags"].split(
                                ","
                            )
                        else:
                            todo.icalendar_component["categories"] = []
                        todo.icalendar_component["related-to"] = task["parent"]
                        todo.icalendar_component["x-errands-color"] = task["color"]
                        todo.save()
                        if task["completed"]:
                            todo.complete()
                        UserData.update_props(
                            calendar.id, task["uid"], ["synced"], [True]
                        )
                    except Exception as e:
                        Log.error(
                            f"Sync: Can't update task on remote: {task['uid']}. {e}"
                        )

                # Delete local task that was deleted on remote
                elif task["uid"] not in remote_ids and task["synced"]:
                    Log.debug(
                        f"Sync: Delete local task deleted on remote: {task['uid']}"
                    )
                    UserData.run_sql(
                        f"""DELETE FROM tasks WHERE uid = '{task["uid"]}'"""
                    )

                # Delete tasks on remote if they were deleted locally
                elif task["uid"] in remote_ids and task["deleted"]:
                    try:
                        Log.debug(f"Sync: Delete task from remote: '{task['uid']}'")
                        if todo := calendar.todo_by_uid(task["uid"]):
                            todo.delete()
                    except Exception as e:
                        Log.error(
                            f"Sync: Can't delete task from remote: '{task['uid']}'. {e}"
                        )

            # Create new local task that was created on CalDAV
            for task in remote_tasks:
                if task["uid"] not in local_ids and task["uid"] not in deleted_uids:
                    Log.debug(
                        f"Sync: Copy new task from remote to list '{calendar.id}': {task['uid']}"
                    )
                    UserData.add_task(
                        color=task["color"],
                        completed=task["completed"],
                        end_date=task["end_date"],
                        list_uid=calendar.id,
                        notes=task["notes"],
                        parent=task["parent"],
                        percent_complete=task["percent_complete"],
                        priority=task["priority"],
                        start_date=task["start_date"],
                        synced=True,
                        tags=task["tags"],
                        text=task["text"],
                        uid=task["uid"],
                    )



# Copyright 2023-2024 Vlad Krupinskii <mrvladus@yandex.ru>
# SPDX-License-Identifier: MIT

import json
import os
import shutil
import sqlite3
from typing import Any, TypedDict
from uuid import uuid4
from errands.lib.gsettings import GSettings
from gi.repository import GLib  # type:ignore
from errands.lib.logging import Log


class TaskListData(TypedDict):
    deleted: bool
    name: str
    synced: bool
    uid: str


class TaskData(TypedDict):
    color: str
    completed: bool
    deleted: bool
    end_date: str
    expanded: bool
    list_uid: str
    notes: str
    parent: str
    percent_complete: int
    priority: int
    start_date: str
    synced: bool
    tags: str
    text: str
    trash: bool
    uid: str


class UserData:
    data_dir: str = os.path.join(GLib.get_user_data_dir(), "errands")
    db_path: str = os.path.join(data_dir, "data.db")

    @classmethod
    def init(cls):
        if not os.path.exists(cls.data_dir):
            os.mkdir(cls.data_dir)
        cls.connection = sqlite3.connect(
            cls.db_path, check_same_thread=False, isolation_level=None
        )
        cls.run_sql(
            """CREATE TABLE IF NOT EXISTS lists (
            deleted INTEGER NOT NULL,
            name TEXT NOT NULL,
            synced INTEGER NOT NULL,
            uid TEXT NOT NULL
            )""",
            """CREATE TABLE IF NOT EXISTS tasks (
            color TEXT NOT NULL,
            completed INTEGER NOT NULL,
            deleted INTEGER NOT NULL,
            end_date TEXT NOT NULL,
            expanded INTEGER NOT NULL,
            list_uid TEXT NOT NULL,
            notes TEXT NOT NULL,
            parent TEXT NOT NULL,
            percent_complete INTEGER NOT NULL,
            priority INTEGER NOT NULL,
            start_date TEXT NOT NULL,
            synced INTEGER NOT NULL,
            tags TEXT NOT NULL,
            text TEXT NOT NULL,
            trash INTEGER NOT NULL,
            uid TEXT NOT NULL
            )""",
        )
        cls._convert(cls)

    @classmethod
    def add_list(cls, name: str, uuid: str = None, synced: bool = False) -> str:
        uid: str = str(uuid4()) if not uuid else uuid
        Log.debug(f"Data: Create '{uid}' list")
        with cls.connection:
            cur = cls.connection.cursor()
            cur.execute(
                "INSERT INTO lists (deleted, name, synced, uid) VALUES (?, ?, ?, ?)",
                (False, name, synced, uid),
            )
            return uid

    @classmethod
    def get_tasks(cls) -> list[str]:
        # Log.debug(f"Data: Get tasks uids")

        with cls.connection:
            cur = cls.connection.cursor()
            cur.execute("SELECT uid FROM tasks")
            return [i[0] for i in cur.fetchall()]

    @classmethod
    def clean_deleted(cls):
        Log.debug("Data: Clean deleted")

        with cls.connection:
            cur = cls.connection.cursor()
            cur.execute("DELETE FROM lists WHERE deleted = 1")
            cur.execute("DELETE FROM tasks WHERE deleted = 1")

    @classmethod
    def get_lists_as_dicts(cls) -> list[TaskListData]:
        # Log.debug("Data: Get lists as dicts")

        with cls.connection:
            cur = cls.connection.cursor()
            cur.execute("SELECT * FROM lists")
            res: list[tuple] = cur.fetchall()
            lists: list[TaskListData] = []
            for i in res:
                lists.append(
                    TaskListData(
                        deleted=bool(i[0]),
                        name=i[1],
                        synced=bool(i[2]),
                        uid=i[3],
                    )
                )

            return lists

    @classmethod
    def move_task_before(cls, list_uid: str, task_uid: str, before_uid: str) -> None:
        tasks: list[TaskData] = UserData.get_tasks_as_dicts()

        # Find tasks to move
        task_to_move_down: TaskData = None
        task_to_move_up: TaskData = None
        for task in tasks:
            if task["list_uid"] == list_uid and task["uid"] == before_uid:
                task_to_move_down = task
            elif task["list_uid"] == list_uid and task["uid"] == task_uid:
                task_to_move_up = task

        # Move task up
        task_to_move_up = tasks.pop(tasks.index(task_to_move_up))
        tasks.insert(tasks.index(task_to_move_down), task_to_move_up)

        # Run SQL
        values: list[tuple[str, str]] = ((t["list_uid"], t["uid"]) for t in tasks)
        with cls.connection:
            cur = cls.connection.cursor()
            cur.execute("CREATE TABLE IF NOT EXISTS tmp AS SELECT * FROM tasks WHERE 0")
            cur.executemany(
                f"""INSERT INTO tmp SELECT * FROM tasks
                    WHERE list_uid = ? AND uid = ?""",
                values,
            )
            cur.execute("DROP TABLE tasks")
            cur.execute("ALTER TABLE tmp RENAME TO tasks")

    @classmethod
    def execute(
        cls, cmd: str, values: tuple = (), fetch: bool = False
    ) -> list[tuple] | None:
        try:
            with cls.connection:
                cur = cls.connection.cursor()
                cur.execute(cmd, values)
                if fetch:
                    return cur.fetchall()
        except Exception as e:
            Log.error(f"Data: {e}")

    @classmethod
    def move_task_to_list(
        cls,
        task_uid: str,
        old_list_uid: str,
        new_list_uid: str,
        parent: str,
        synced: bool,
    ) -> None:
        sub_tasks_uids: list[str] = cls.get_tasks_uids_tree(old_list_uid, task_uid)
        tasks: list[TaskData] = cls.get_tasks_as_dicts(old_list_uid)
        task_dict: TaskData = [i for i in tasks if i["uid"] == task_uid][0]
        sub_tasks_dicts: list[TaskData] = [
            i for i in tasks if i["uid"] in sub_tasks_uids
        ]
        # Move task
        if old_list_uid == new_list_uid:
            cls.run_sql(
                f"DELETE FROM tasks WHERE list_uid = '{old_list_uid}' AND uid = '{task_uid}'"
            )
        else:
            cls.update_props(old_list_uid, task_uid, ["deleted"], [True])
        task_dict["list_uid"] = new_list_uid
        task_dict["parent"] = parent
        task_dict["synced"] = synced
        cls.add_task(**task_dict)
        # Move sub-tasks
        for sub_dict in sub_tasks_dicts:
            if old_list_uid == new_list_uid:
                cls.run_sql(
                    f"""DELETE FROM tasks
                    WHERE list_uid = '{old_list_uid}'
                    AND uid = '{sub_dict['uid']}'"""
                )
            else:
                cls.update_props(old_list_uid, sub_dict["uid"], ["deleted"], [True])
            sub_dict["list_uid"] = new_list_uid
            sub_dict["synced"] = synced
            cls.add_task(**sub_dict)

    @classmethod
    def get_prop(cls, list_uid: str, uid: str, prop: str) -> Any:
        # Log.debug(f"Data: Get '{prop}' for '{uid}'")

        with cls.connection:
            cur = cls.connection.cursor()
            cur.execute(
                f"SELECT {prop} FROM tasks WHERE uid = ? AND list_uid = ?",
                (uid, list_uid),
            )
            return cur.fetchone()[0]

    @classmethod
    def update_props(
        cls, list_uid: str, uid: str, props: list[str], values: list[Any]
    ) -> None:
        # Log.debug(f"Data: Update props for '{uid}'")

        with cls.connection:
            cur = cls.connection.cursor()
            for i, prop in enumerate(props):
                cur.execute(
                    f"""UPDATE tasks SET {prop} = ? 
                    WHERE uid = '{uid}'
                    AND list_uid = '{list_uid}'""",
                    (values[i],),
                )

    @classmethod
    def run_sql(cls, *cmds: list, fetch: bool = False) -> list[tuple] | None:
        try:
            with cls.connection:
                cur = cls.connection.cursor()
                for cmd in cmds:
                    if isinstance(cmd, tuple):
                        cur.execute(cmd[0], cmd[1])
                    else:
                        cur.execute(cmd)
                return cur.fetchall() if fetch else None
        except Exception as e:
            Log.error(f"Data: {e}")

    @classmethod
    def get_tasks_uids(cls, list_uid: str, parent: str = None) -> list[str]:
        # Log.debug(f"Data: Get tasks uids {f'for {parent}' if parent else ''}")

        with cls.connection:
            cur = cls.connection.cursor()
            cur.execute(
                f"""SELECT uid FROM tasks
                WHERE list_uid = ?
                AND deleted = 0
                {f"AND parent = '{parent}'" if parent != None else ''}""",
                (list_uid,),
            )
            return [i[0] for i in cur.fetchall()]

    @classmethod
    def get_tasks_uids_tree(cls, list_uid: str, parent: str) -> list[str]:
        """
        Get all sub-task uids recursively
        """
        # Log.debug(f"Data: Get tasks uids tree for '{parent}'")

        uids: list[str] = []

        def _add(sub_uids: list[str]) -> None:
            for uid in sub_uids:
                uids.append(uid)
                if len(cls.get_tasks_uids(list_uid, uid)) > 0:
                    _add(cls.get_tasks_uids(list_uid, uid))

        _add(cls.get_tasks_uids(list_uid, parent))

        return uids

    @classmethod
    def get_tasks_as_dicts(
        cls, list_uid: str = None, parent: str = None
    ) -> list[TaskData]:
        # Log.debug(f"Data: Get tasks as dicts")

        with cls.connection:
            cur = cls.connection.cursor()
            if not list_uid and not parent:
                cur.execute("SELECT * FROM tasks")
            else:
                cur.execute(
                    f"""SELECT * FROM tasks WHERE list_uid = ?
                    {f"AND parent = '{parent}'" if parent else ""}""",
                    (list_uid,),
                )
            tasks: list[TaskData] = []
            for task in cur.fetchall():
                tasks.append(
                    TaskData(
                        color=task[0],
                        completed=bool(task[1]),
                        deleted=bool(task[2]),
                        end_date=task[3],
                        expanded=bool(task[4]),
                        list_uid=task[5],
                        notes=task[6],
                        parent=task[7],
                        percent_complete=int(task[8]),
                        priority=int(task[9]),
                        start_date=task[10],
                        synced=bool(task[11]),
                        tags=task[12],
                        text=task[13],
                        trash=bool(task[14]),
                        uid=task[15],
                    )
                )

            return tasks

    @classmethod
    def add_task(
        cls,
        color: str = "",
        completed: bool = False,
        deleted: bool = False,
        end_date: str = "",
        expanded: bool = False,
        list_uid: str = "",
        notes: str = "",
        parent: str = "",
        percent_complete: int = 0,
        priority: int = 0,
        start_date: str = "",
        synced: bool = False,
        tags: str = "",
        text: str = "",
        trash: bool = False,
        uid: str = "",
    ) -> str:
        if not uid:
            uid: str = str(uuid4())

        Log.debug(f"Data: Add task {uid}")

        with cls.connection:
            cur = cls.connection.cursor()
            cur.execute(
                """INSERT INTO tasks 
                (uid, list_uid, text, parent, completed, deleted, color, notes, percent_complete, priority, start_date, end_date, tags, synced, expanded, trash) 
                VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)""",
                (
                    uid,
                    list_uid,
                    text,
                    parent,
                    completed,
                    deleted,
                    color,
                    notes,
                    percent_complete,
                    priority,
                    start_date,
                    end_date,
                    tags,
                    synced,
                    expanded,
                    trash,
                ),
            )

            return uid

    def _convert(cls):
        old_path = os.path.join(GLib.get_user_data_dir(), "list")
        old_data_file = os.path.join(old_path, "data.json")
        if not os.path.exists(old_data_file):
            return
        Log.debug("Data: convert data file")
        # Get tasks
        try:
            with open(old_data_file, "r") as f:
                data: dict = json.loads(f.read())
        except:
            Log.error("Data: can't read data file")
            return
        # Remove old data folder
        shutil.rmtree(old_path, True)
        # If sync is enabled
        if GSettings.get("sync-provider") != 0:
            uid = cls.add_list(GSettings.get("sync-cal-name"), synced=True)
            GSettings.set("sync-cal-name", "s", "")
        # If sync is disabled
        else:
            uid = cls.add_list("Errands")
        # Add tasks
        for task in data["tasks"]:
            cls.add_task(
                color=task["color"],
                completed=task["completed"],
                deleted=task["id"] in data["deleted"],
                list_uid=uid,
                parent=task["parent"],
                synced=task["synced_caldav"],
                text=task["text"],
                trash=task["deleted"],
                uid=task["id"],
            )



from errands.lib.gsettings import GSettings
from errands.lib.logging import Log
from errands.lib.sync.providers.caldav import SyncProviderCalDAV


class SyncProviderNextcloud(SyncProviderCalDAV):
    def __init__(self, *args, **kwargs) -> None:
        return super().__init__(name="Nextcloud", *args, **kwargs)

    def _check_url(self) -> None:
        Log.debug("Sync: Checking URL")

        # Add prefix if needed
        if not self.url.startswith("http"):
            self.url = "https://" + self.url
            GSettings.set("sync-url", "s", self.url)
        # Add suffix if needed
        if not "remote.php/dav" in GSettings.get("sync-url"):
            self.url = f"{self.url}/remote.php/dav/"
            GSettings.set("sync-url", "s", self.url)
        else:
            self.url = GSettings.get("sync-url")

        Log.debug(f"Sync: URL is set to {self.url}")



# Copyright 2023 Vlad Krupinskii <mrvladus@yandex.ru>
# SPDX-License-Identifier: MIT

from __future__ import annotations
import time
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from errands.widgets.window import Window

from datetime import datetime
from icalendar import Calendar, Todo
from errands.lib.data import UserData
from errands.lib.functions import get_children
from errands.lib.gsettings import GSettings
from errands.lib.logging import Log
from errands.lib.sync.sync import Sync
from errands.widgets.components import Box
from errands.widgets.task import Task
from errands.widgets.task_list import TaskList
from gi.repository import Adw, Gtk, Gio, GObject, Gdk, GLib  # type:ignore


class Sidebar(Adw.Bin):
    def __init__(self, window: Window):
        super().__init__()
        self.window: Window = window
        self._build_ui()

    def _build_ui(self):
        # Components
        self.status_page: SidebarStatusPage = SidebarStatusPage(self)
        self.task_lists: SidebarTaskLists = SidebarTaskLists(self)
        self.header_bar: SidebarHeaderBar = SidebarHeaderBar(
            self.task_lists, self.window
        )
        self.trash_button: SidebarTrashButton = SidebarTrashButton(
            self.task_lists, self.window
        )
        # Toolbar view
        toolbar_view: Adw.ToolbarView = Adw.ToolbarView(
            content=Box(
                children=[
                    self.status_page,
                    self.task_lists,
                    self.trash_button,
                ],
                orientation=Gtk.Orientation.VERTICAL,
            )
        )
        toolbar_view.add_top_bar(self.header_bar)
        self.set_child(toolbar_view)


class SidebarHeaderBar(Adw.Bin):
    def __init__(self, task_lists: SidebarTaskLists, window: Window):
        super().__init__()
        self.task_lists: SidebarTaskLists = task_lists
        self.window: Window = window
        self._build_ui()

    def _build_ui(self):
        # HeaderBar
        hb: Adw.HeaderBar = Adw.HeaderBar(
            title_widget=Gtk.Label(
                label=_("Errands"),
                css_classes=["heading"],
            )
        )
        self.set_child(hb)

        # Import menu
        import_menu: Gio.Menu = Gio.Menu.new()
        import_menu.append(_("Import List"), "app.import")

        # Add list button
        self.add_list_btn: Adw.SplitButton = Adw.SplitButton(
            icon_name="list-add-symbolic",
            tooltip_text=_("Add List (Ctrl+A)"),
            menu_model=import_menu,
            dropdown_tooltip=_("More Options"),
        )
        ctrl = Gtk.ShortcutController(scope=1)
        ctrl.add_shortcut(
            Gtk.Shortcut(
                trigger=Gtk.ShortcutTrigger.parse_string("<Primary>A"),
                action=Gtk.ShortcutAction.parse_string("activate"),
            )
        )
        self.add_list_btn.add_controller(ctrl)
        self.add_list_btn.connect("clicked", self._on_add_btn_clicked)
        hb.pack_start(self.add_list_btn)

        # Main menu
        menu: Gio.Menu = Gio.Menu.new()
        top_section: Gio.Menu = Gio.Menu.new()
        # top_section.append(_("Secret Notes"), "app.secret_notes")
        top_section.append(_("Sync / Fetch Tasks"), "app.sync")
        menu.append_section(None, top_section)
        bottom_section: Gio.Menu = Gio.Menu.new()
        bottom_section.append(_("Preferences"), "app.preferences")
        bottom_section.append(_("Keyboard Shortcuts"), "win.show-help-overlay")
        bottom_section.append(_("About Errands"), "app.about")
        menu.append_section(None, bottom_section)
        menu_btn: Gtk.MenuButton = Gtk.MenuButton(
            menu_model=menu,
            primary=True,
            icon_name="open-menu-symbolic",
            tooltip_text=_("Main Menu"),
        )
        hb.pack_end(menu_btn)

        # Sync indicator
        self.sync_indicator: Gtk.Spinner = Gtk.Spinner(
            tooltip_text=_("Syncing..."), visible=False, spinning=True
        )
        hb.pack_end(self.sync_indicator)

    def _on_add_btn_clicked(self, btn) -> None:
        def _entry_activated(_, dialog):
            if dialog.get_response_enabled("add"):
                dialog.response("add")
                dialog.close()

        def _entry_changed(entry, _, dialog):
            text = entry.props.text.strip(" \n\t")
            names = [i["name"] for i in UserData.get_lists_as_dicts()]
            dialog.set_response_enabled("add", text and text not in names)

        def _confirm(_, res, entry):
            if res == "cancel":
                return

            name = entry.props.text.rstrip().lstrip()
            uid = UserData.add_list(name)
            row = self.task_lists.add_list(name, uid)
            self.task_lists.lists.select_row(row)
            Sync.sync()

        entry = Gtk.Entry(placeholder_text=_("New List Name"))
        dialog = Adw.MessageDialog(
            transient_for=self.window,
            hide_on_close=True,
            heading=_("Add List"),
            default_response="add",
            close_response="cancel",
            extra_child=entry,
        )
        dialog.add_response("cancel", _("Cancel"))
        dialog.add_response("add", _("Add"))
        dialog.set_response_enabled("add", False)
        dialog.set_response_appearance("add", Adw.ResponseAppearance.SUGGESTED)
        dialog.connect("response", _confirm, entry)
        entry.connect("activate", _entry_activated, dialog)
        entry.connect("notify::text", _entry_changed, dialog)
        dialog.present()


class SidebarStatusPage(Adw.Bin):
    def __init__(self, sidebar: Sidebar):
        super().__init__()
        self.sidebar: Sidebar = sidebar
        self._build_ui()

    def _build_ui(self) -> None:
        status_page: Adw.StatusPage = Adw.StatusPage(
            title=_("Add new List"),
            description=_('Click "+" button'),
            icon_name="errands-lists-symbolic",
            css_classes=["compact"],
            vexpand=True,
        )
        self.set_child(status_page)


class SidebarTaskLists(Gtk.ScrolledWindow):
    def __init__(self, sidebar: Sidebar):
        super().__init__()
        self.sidebar: Sidebar = sidebar
        self.window: Window = sidebar.window
        self._build_ui()
        self._load_lists()
        self.update_ui()

    def _build_ui(self) -> None:
        self.set_propagate_natural_height(True)
        self.set_vexpand(True)
        # Lists
        self.lists: Gtk.ListBox = Gtk.ListBox(css_classes=["navigation-sidebar"])
        self.lists.connect("row-selected", self._on_row_selected)
        self.bind_property(
            "visible",
            self.sidebar.status_page,
            "visible",
            GObject.BindingFlags.SYNC_CREATE
            | GObject.BindingFlags.INVERT_BOOLEAN
            | GObject.BindingFlags.BIDIRECTIONAL,
        )
        self.set_child(self.lists)

    def add_list(self, name, uid) -> SidebarTaskListsItem:
        task_list: TaskList = TaskList(self.window, uid, self)
        page: Adw.ViewStackPage = self.window.stack.add_titled(
            child=task_list, name=name, title=name
        )
        task_list.title.bind_property(
            "title", page, "title", GObject.BindingFlags.SYNC_CREATE
        )
        task_list.title.bind_property(
            "title", page, "name", GObject.BindingFlags.SYNC_CREATE
        )
        row: SidebarTaskListsItem = SidebarTaskListsItem(task_list, self)
        self.lists.append(row)
        return row

    def _on_row_selected(self, _, row: Gtk.ListBoxRow) -> None:
        Log.debug("Lists: Switch list")
        if row:
            name = row.label.get_label()
            self.window.stack.set_visible_child_name(name)
            self.window.split_view.set_show_content(True)
            GSettings.set("last-open-list", "s", name)
            self.sidebar.status_page.set_visible(False)

    def _get_task_lists(self) -> list[TaskList]:
        lists: list[TaskList] = []
        pages: Adw.ViewStackPages = self.window.stack.get_pages()
        for i in range(pages.get_n_items()):
            child: TaskList = pages.get_item(i).get_child()
            if isinstance(child, TaskList):
                lists.append(child)
        return lists

    def _get_task_lists_items(self) -> list[SidebarTaskListsItem]:
        return get_children(self.lists)

    def _load_lists(self) -> None:
        lists: list[dict] = [
            i for i in UserData.get_lists_as_dicts() if not i["deleted"]
        ]
        for l in lists:
            row: SidebarTaskListsItem = self.add_list(l["name"], l["uid"])
            if GSettings.get("last-open-list") == l["name"]:
                self.lists.select_row(row)
        self.sidebar.status_page.set_visible(len(lists) == 0)

    def update_ui(self) -> None:
        Log.debug("Lists: Update UI...")

        # Delete lists
        uids: list[str] = [i["uid"] for i in UserData.get_lists_as_dicts()]
        for row in self._get_task_lists_items():
            if row.uid not in uids:
                prev_child: SidebarTaskListsItem | None = row.get_prev_sibling()
                next_child: SidebarTaskListsItem | None = row.get_next_sibling()
                list: TaskList = row.task_list
                self.window.stack.remove(list)
                if prev_child or next_child:
                    self.lists.select_row(prev_child or next_child)
                self.lists.remove(row)

        # Update old lists
        for l in self._get_task_lists():
            l.update_ui()

        # Create new lists
        old_uids: list[str] = [row.uid for row in self._get_task_lists_items()]
        new_lists: list[dict] = UserData.get_lists_as_dicts()
        for l in new_lists:
            if l["uid"] not in old_uids:
                Log.debug(f"Lists: Add list '{l['uid']}'")
                row: SidebarTaskListsItem = self.add_list(l["name"], l["uid"])
                self.lists.select_row(row)
                self.window.stack.set_visible_child_name(l["name"])
                self.sidebar.status_page.set_visible(False)

        # Show status
        length: int = len(self._get_task_lists_items())
        self.sidebar.status_page.set_visible(length == 0)
        if length == 0:
            self.window.stack.set_visible_child_name("status")


class SidebarTaskListsItem(Gtk.ListBoxRow):
    def __init__(self, task_list: TaskList, task_lists: SidebarTaskLists) -> None:
        super().__init__()
        self.task_lists: SidebarTaskLists = task_lists
        self.list_box: Gtk.ListBox = task_lists.lists
        self.task_list: TaskList = task_list
        self.uid: str = task_list.list_uid
        self.window: Window = task_lists.window
        self._build_ui()
        self._add_actions()

    def _add_actions(self) -> None:
        group: Gio.SimpleActionGroup = Gio.SimpleActionGroup()
        self.insert_action_group(name="list_item", group=group)

        def _create_action(name: str, callback: callable) -> None:
            action: Gio.SimpleAction = Gio.SimpleAction.new(name, None)
            action.connect("activate", callback)
            group.add_action(action)

        def _delete(*args):
            def _confirm(_, res):
                if res == "cancel":
                    Log.debug("ListItem: Deleting list is cancelled")
                    return

                Log.info(f"Lists: Delete list '{self.uid}'")
                UserData.run_sql(
                    f"UPDATE lists SET deleted = 1 WHERE uid = '{self.uid}'",
                    f"DELETE FROM tasks WHERE list_uid = '{self.uid}'",
                )
                self.window.stack.remove(self.task_list)
                # Switch row
                next_row: SidebarTaskListsItem = self.get_next_sibling()
                prev_row: SidebarTaskListsItem = self.get_prev_sibling()
                self.list_box.remove(self)
                if next_row or prev_row:
                    self.list_box.select_row(next_row or prev_row)
                else:
                    self.window.stack.set_visible_child_name("status")
                    self.task_lists.sidebar.status_page.set_visible(True)

                # Remove items from trash
                for item in get_children(self.window.trash.trash_list):
                    if item.task_widget.list_uid == self.uid:
                        self.window.trash.trash_list.remove(item)

                Sync.sync()

            dialog: Adw.MessageDialog = Adw.MessageDialog(
                transient_for=self.window,
                hide_on_close=True,
                heading=_("Are you sure?"),
                body=_("List will be permanently deleted"),
                default_response="delete",
                close_response="cancel",
            )
            dialog.add_response("cancel", _("Cancel"))
            dialog.add_response("delete", _("Delete"))
            dialog.set_response_appearance("delete", Adw.ResponseAppearance.DESTRUCTIVE)
            dialog.connect("response", _confirm)
            dialog.present()

        def _rename(*args):
            def _entry_activated(_, dialog: Adw.MessageDialog):
                if dialog.get_response_enabled("save"):
                    dialog.response("save")
                    dialog.close()

            def _entry_changed(entry: Gtk.Entry, _, dialog: Adw.MessageDialog):
                text = entry.props.text.strip(" \n\t")
                names = [i["name"] for i in UserData.get_lists_as_dicts()]
                dialog.set_response_enabled("save", text and text not in names)

            def _confirm(_, res, entry: Gtk.Entry):
                if res == "cancel":
                    Log.debug("ListItem: Editing list name is cancelled")
                    return
                Log.info(f"ListItem: Rename list {self.uid}")

                text: str = entry.props.text.rstrip().lstrip()
                UserData.run_sql(
                    (
                        "UPDATE lists SET name = ?, synced = 0 WHERE uid = ?",
                        (text, self.uid),
                    )
                )
                self.task_list.title.set_title(text)
                Sync.sync()

            entry: Gtk.Entry = Gtk.Entry(placeholder_text=_("New Name"))
            entry.get_buffer().props.text = self.label.get_label()
            dialog: Adw.MessageDialog = Adw.MessageDialog(
                transient_for=self.window,
                hide_on_close=True,
                heading=_("Rename List"),
                default_response="save",
                close_response="cancel",
                extra_child=entry,
            )
            dialog.add_response("cancel", _("Cancel"))
            dialog.add_response("save", _("Save"))
            dialog.set_response_enabled("save", False)
            dialog.set_response_appearance("save", Adw.ResponseAppearance.SUGGESTED)
            dialog.connect("response", _confirm, entry)
            entry.connect("activate", _entry_activated, dialog)
            entry.connect("notify::text", _entry_changed, dialog)
            dialog.present()

        def _export(*args):
            def _confirm(dialog, res):
                try:
                    file = dialog.save_finish(res)
                except:
                    Log.debug("List: Export cancelled")
                    return

                Log.info(f"List: Export '{self.uid}'")

                tasks: list[dict] = UserData.get_tasks_as_dicts(self.uid)
                calendar: Calendar = Calendar()
                calendar.add("x-wr-calname", self.label.get_label())
                for task in tasks:
                    event = Todo()
                    event.add("uid", task["uid"])
                    event.add("related-to", task["parent"])
                    event.add("summary", task["text"])
                    if task["notes"]:
                        event.add("description", task["notes"])
                    event.add("priority", task["priority"])
                    if task["tags"]:
                        event.add("categories", task["tags"])
                    event.add("percent-complete", task["percent_complete"])
                    if task["color"]:
                        event.add("x-errands-color", task["color"])
                    event.add(
                        "dtstart",
                        (
                            datetime.fromisoformat(task["start_date"])
                            if task["start_date"]
                            else datetime.now()
                        ),
                    )
                    if task["end_date"]:
                        event.add(
                            "due",
                            (
                                datetime.fromisoformat(task["end_date"])
                                if task["end_date"]
                                else datetime.now()
                            ),
                        )
                    calendar.add_component(event)

                try:
                    with open(file.get_path(), "wb") as f:
                        f.write(calendar.to_ical())
                except Exception as e:
                    Log.error(f"List: Export failed. {e}")
                    self.window.add_toast(_("Export failed"))

                self.window.add_toast(_("Exported"))

            filter: Gtk.FileFilter = Gtk.FileFilter()
            filter.add_pattern("*.ics")
            dialog: Gtk.FileDialog = Gtk.FileDialog(
                initial_name=f"{self.uid}.ics", default_filter=filter
            )
            dialog.save(self.window, None, _confirm)

        _create_action("delete", _delete)
        _create_action("rename", _rename)
        _create_action("export", _export)

    def _build_ui(self) -> None:
        self.add_css_class("task-lists-item")

        # Label
        self.label: Gtk.Label = Gtk.Label(
            halign="start",
            hexpand=True,
            ellipsize=3,
        )
        self.task_list.title.bind_property(
            "title",
            self.label,
            "label",
            GObject.BindingFlags.SYNC_CREATE,
        )

        # Menu
        menu: Gio.Menu = Gio.Menu.new()
        menu.append(_("Rename"), "list_item.rename")
        menu.append(_("Delete"), "list_item.delete")
        menu.append(_("Export"), "list_item.export")

        # Click controller
        ctrl: Gtk.GestureClick = Gtk.GestureClick()
        ctrl.connect("released", self._on_click)
        self.add_controller(ctrl)

        # Drop controller
        drop_ctrl: Gtk.DropTarget = Gtk.DropTarget.new(
            actions=Gdk.DragAction.MOVE, type=Task
        )
        drop_ctrl.connect("drop", self._on_task_drop)
        self.add_controller(drop_ctrl)

        # Drop hover controller
        drop_hover_ctrl: Gtk.DropControllerMotion = Gtk.DropControllerMotion()
        drop_hover_ctrl.connect("enter", self._on_drop_hover)
        self.add_controller(drop_hover_ctrl)

        self.set_child(
            Box(
                children=[
                    self.label,
                    Gtk.MenuButton(
                        menu_model=menu,
                        icon_name="view-more-symbolic",
                        tooltip_text=_("Menu"),
                    ),
                ],
                css_classes=["toolbar"],
            )
        )

    def _on_click(self, *args) -> None:
        self.window.stack.set_visible_child_name(self.label.get_label())
        self.window.split_view.set_show_content(True)

    def _on_drop_hover(self, ctrl: Gtk.DropControllerMotion, _x, _y):
        """
        Switch list on dnd hover after DELAY_SECONDS
        """

        DELAY_SECONDS: float = 0.7
        entered_at: float = time.time()

        def _switch_delay():
            if ctrl.contains_pointer():
                if time.time() - entered_at >= DELAY_SECONDS:
                    self.activate()
                    return False
                else:
                    return True
            else:
                return False

        GLib.timeout_add(100, _switch_delay)

    def _on_task_drop(self, _drop, task: Task, _x, _y):
        """
        Move task and sub-tasks to new list
        """

        if task.list_uid == self.uid:
            return

        Log.info(f"Lists: Move '{task.uid}' to '{self.uid}' list")
        UserData.move_task_to_list(task.uid, task.list_uid, self.uid, "", False)
        uid: str = task.uid
        task.purge()
        self.task_list.add_task(uid)
        Sync.sync()


class SidebarTrashButton(Gtk.Button):
    def __init__(self, task_lists: SidebarTaskLists, window: Window):
        super().__init__()
        self.task_lists: SidebarTaskLists = task_lists
        self.window: Window = window
        self._build_ui()

    def _build_ui(self):
        self.set_child(
            Adw.ButtonContent(
                icon_name="errands-trash-symbolic",
                label=_("Trash"),
                halign="center",
            )
        )
        self.add_css_class("flat")
        self.set_margin_top(6)
        self.set_margin_bottom(6)
        self.set_margin_start(6)
        self.set_margin_end(6)
        self.task_lists.bind_property(
            "visible",
            self,
            "visible",
            GObject.BindingFlags.SYNC_CREATE,
        )

        # Drop controller
        trash_drop_ctrl = Gtk.DropTarget.new(actions=Gdk.DragAction.MOVE, type=Task)
        trash_drop_ctrl.connect("drop", lambda _d, t, _x, _y: t.delete())
        self.add_controller(trash_drop_ctrl)

    def do_clicked(self) -> None:
        self.task_lists.lists.unselect_all()
        self.window.stack.set_visible_child_name("trash")
        self.window.split_view.set_show_content(True)



# Copyright 2023-2024 Vlad Krupinskii <mrvladus@yandex.ru>
# SPDX-License-Identifier: MIT

from unittest import TestCase
from errands.widgets.window import WINDOW


class TestTaskList(TestCase):
    def setUp(self):
        self.window = WINDOW



# Copyright 2023 Vlad Krupinskii <mrvladus@yandex.ru>
# SPDX-License-Identifier: MIT

from __future__ import annotations
from __main__ import VERSION, APP_ID
from uuid import uuid4
from icalendar import Calendar
from errands.lib.data import UserData
from errands.widgets.components import Box, Button
from errands.widgets.secret_notes import SecretNotesWindow
from errands.widgets.trash import Trash
from gi.repository import Gio, Adw, Gtk  # type:ignore
from errands.widgets.sidebar import Sidebar
from errands.widgets.preferences import PreferencesWindow
from errands.lib.sync.sync import Sync
from errands.lib.gsettings import GSettings
from errands.lib.logging import Log

WINDOW: Window = None


class Window(Adw.ApplicationWindow):
    about_window: Adw.AboutWindow = None
    secret_notes: SecretNotesWindow = None

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        global WINDOW
        WINDOW = self
        self._create_actions()
        self._build_ui()
        self.present()
        # Setup sync
        Sync.window = self
        Sync.sync()

    def _build_ui(self):
        self.set_title(_("Errands"))
        self.props.width_request = 360
        self.props.height_request = 200
        # Remember window state
        GSettings.bind("width", self, "default_width")
        GSettings.bind("height", self, "default_height")
        GSettings.bind("maximized", self, "maximized")
        # Setup theme
        Adw.StyleManager.get_default().set_color_scheme(GSettings.get("theme"))

        # Split View
        self.split_view = Adw.NavigationSplitView(
            max_sidebar_width=300,
            min_sidebar_width=240,
            show_content=True,
            sidebar_width_fraction=0.25,
        )

        # Stack
        self.stack = Adw.ViewStack()

        # Trash
        self.trash = Trash(self)
        self.stack.add_titled(self.trash, name="trash", title=_("Trash"))

        # Status page Toolbar View
        status_toolbar_view = Adw.ToolbarView(
            content=Box(
                children=[
                    Adw.StatusPage(title=_("No Task Lists"), icon_name=APP_ID),
                    Button(
                        label=_("Create List"),
                        on_click=lambda *_: self.sidebar.header_bar.add_list_btn.activate(),
                        halign="center",
                        css_classes=["pill", "suggested-action"],
                    ),
                ],
                orientation="vertical",
                vexpand=True,
                valign="center",
            )
        )
        status_toolbar_view.add_top_bar(Adw.HeaderBar(show_title=False))
        self.stack.add_titled(
            child=status_toolbar_view,
            name="status",
            title=_("No Task Lists"),
        )
        self.stack.set_visible_child_name("status")

        # Lists
        self.sidebar = Sidebar(self)
        self.split_view.set_sidebar(Adw.NavigationPage.new(self.sidebar, _("Lists")))
        self.split_view.set_content(Adw.NavigationPage.new(self.stack, _("Tasks")))

        # Toast overlay
        self.toast_overlay = Adw.ToastOverlay(child=self.split_view)

        # Breakpoints
        bp = Adw.Breakpoint.new(Adw.breakpoint_condition_parse("max-width: 980px"))
        bp.add_setter(self.split_view, "collapsed", True)
        bp.add_setter(self.split_view, "show-content", True)
        self.add_breakpoint(bp)

        self.set_content(self.toast_overlay)

    def add_toast(self, text: str) -> None:
        self.toast_overlay.add_toast(Adw.Toast.new(title=text))

    def _create_actions(self) -> None:
        """
        Create actions for main menu
        """
        Log.debug("Creating actions")

        def _create_action(name: str, callback: callable, shortcuts=None) -> None:
            action: Gio.SimpleAction = Gio.SimpleAction.new(name, None)
            action.connect("activate", callback)
            if shortcuts:
                self.props.application.set_accels_for_action(f"app.{name}", shortcuts)
            self.props.application.add_action(action)

        def _about(*args) -> None:
            """
            Show about window
            """
            if not self.about_window:
                self.about_window = Adw.AboutWindow(
                    transient_for=self,
                    version=VERSION,
                    application_icon=APP_ID,
                    application_name=_("Errands"),
                    copyright="© 2023 Vlad Krupinskii",
                    website="https://github.com/mrvladus/Errands",
                    issue_url="https://github.com/mrvladus/Errands/issues",
                    license_type=Gtk.License.MIT_X11,
                    translator_credits=_("translator-credits"),
                    modal=True,
                    hide_on_close=True,
                )
            self.about_window.present()

        def _sync(*args):
            Sync.sync()
            if GSettings.get("sync-provider") == 0:
                self.add_toast(_("Sync is disabled"))

        def _import(*args) -> None:
            def _confirm(dialog: Gtk.FileDialog, res) -> None:
                try:
                    file: Gio.File = dialog.open_finish(res)
                except:
                    Log.debug("Lists: Import cancelled")
                    return
                with open(file.get_path(), "r") as f:
                    calendar: Calendar = Calendar.from_ical(f.read())
                    # List name
                    name = calendar.get(
                        "X-WR-CALNAME", file.get_basename().rstrip(".ics")
                    )
                    if name in [
                        i[0]
                        for i in UserData.run_sql("SELECT name FROM lists", fetch=True)
                    ]:
                        name = f"{name}_{uuid4()}"
                    # Create list
                    uid: str = UserData.add_list(name)
                    # Add tasks
                    for todo in calendar.walk("VTODO"):
                        # Tags
                        if (tags := todo.get("CATEGORIES", "")) != "":
                            tags = ",".join(
                                [
                                    i.to_ical().decode("utf-8")
                                    for i in (
                                        tags if isinstance(tags, list) else tags.cats
                                    )
                                ]
                            )
                        # Start
                        if (start := todo.get("DTSTART", "")) != "":
                            start = (
                                todo.get("DTSTART", "")
                                .to_ical()
                                .decode("utf-8")
                                .strip("Z")
                            )
                        else:
                            start = ""
                        # End
                        if (end := todo.get("DUE", todo.get("DTEND", ""))) != "":
                            end = (
                                todo.get("DUE", todo.get("DTEND", ""))
                                .to_ical()
                                .decode("utf-8")
                                .strip("Z")
                            )
                        else:
                            end = ""
                        UserData.add_task(
                            color=todo.get("X-ERRANDS-COLOR", ""),
                            completed=str(todo.get("STATUS", "")) == "COMPLETED",
                            end_date=end,
                            list_uid=uid,
                            notes=str(todo.get("DESCRIPTION", "")),
                            parent=str(todo.get("RELATED-TO", "")),
                            percent_complete=int(todo.get("PERCENT-COMPLETE", 0)),
                            priority=int(todo.get("PRIORITY", 0)),
                            start_date=start,
                            tags=tags,
                            text=str(todo.get("SUMMARY", "")),
                            uid=todo.get("UID", None),
                        )
                self.sidebar.task_lists.update_ui()
                self.add_toast(_("Imported"))
                Sync.sync()

            filter = Gtk.FileFilter()
            filter.add_pattern("*.ics")
            dialog = Gtk.FileDialog(default_filter=filter)
            dialog.open(self, None, _confirm)

        def _secret_notes(*args):
            if not self.secret_notes:
                self.secret_notes: SecretNotesWindow = SecretNotesWindow(self)
                self.secret_notes.present()
            else:
                self.secret_notes.present()

        _create_action(
            "preferences",
            lambda *_: PreferencesWindow(self).show(),
            ["<primary>comma"],
        )
        _create_action("about", _about)
        _create_action("import", _import)
        _create_action("secret_notes", _secret_notes)
        _create_action("sync", _sync, ["<primary>f"])
        _create_action(
            "quit",
            lambda *_: self.props.application.quit(),
            ["<primary>q", "<primary>w"],
        )



# Copyright 2023 Vlad Krupinskii <mrvladus@yandex.ru>
# SPDX-License-Identifier: MIT

import re
from gi.repository import GLib  # type:ignore


class Markup:
    @classmethod
    def escape(self, text: str) -> str:
        return GLib.markup_escape_text(text)

    @classmethod
    def find_url(self, text: str) -> str:
        """Convert urls to markup. Make sure to escape text before calling."""

        string: str = text
        urls: list[str] = re.findall(r"(https?://\S+)", string)
        for url in urls:
            string = string.replace(url, f'<a href="{url}">{url}</a>')
        return string



# Copyright 2023 Vlad Krupinskii <mrvladus@yandex.ru>
# SPDX-License-Identifier: MIT

from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from errands.widgets.window import Window

from errands.lib.sync.providers.caldav import SyncProviderCalDAV
from errands.lib.sync.providers.nextcloud import SyncProviderNextcloud
from errands.lib.gsettings import GSettings
from errands.lib.logging import Log
from errands.lib.data import UserData
from errands.lib.functions import threaded
from gi.repository import GLib  # type:ignore


class Sync:
    provider = None
    window: Window = None
    syncing: bool = False
    sync_again: bool = False

    @classmethod
    def init(self, window: Window, testing: bool = False) -> None:
        self.window: Window = window
        Log.info("Sync: Initialize sync provider")
        match GSettings.get("sync-provider"):
            case 0:
                Log.info("Sync: Sync disabled")
                UserData.clean_deleted()
            case 1:
                self.provider = SyncProviderNextcloud(
                    window=self.window, testing=testing
                )
            case 2:
                self.provider = SyncProviderCalDAV(window=self.window, testing=testing)

    @classmethod
    @threaded
    def sync(self, update_ui: bool = True) -> None:
        """
        Sync tasks without blocking the UI
        """
        if GSettings.get("sync-provider") == 0:
            UserData.clean_deleted()
            if update_ui:
                GLib.idle_add(self.window.sidebar.task_lists.update_ui)
            return
        if not self.provider:
            GLib.idle_add(
                self.window.sidebar.header_bar.sync_indicator.set_visible, True
            )
            self.init(self.window)
            GLib.idle_add(
                self.window.sidebar.header_bar.sync_indicator.set_visible, False
            )
        if self.provider and self.provider.can_sync:
            if self.syncing:
                self.sync_again = True
                return
            self.syncing = True
            GLib.idle_add(
                self.window.sidebar.header_bar.sync_indicator.set_visible, True
            )
            self.provider.sync()
            UserData.clean_deleted()
            if update_ui:
                GLib.idle_add(self.window.sidebar.task_lists.update_ui)
            if self.sync_again:
                self.sync()
                self.sync_again = False
            GLib.idle_add(
                self.window.sidebar.header_bar.sync_indicator.set_visible, False
            )
            self.syncing = False

    @classmethod
    def test_connection(self) -> bool:
        self.init(testing=True, window=self.window)
        return self.provider.can_sync



# Copyright 2023-2024 Vlad Krupinskii <mrvladus@yandex.ru>
# SPDX-License-Identifier: MIT

from unittest import TestCase
from gi.repository import Adw, Gio

from errands.widgets.window import WINDOW


class TestWindow(TestCase):
    def setUp(self):
        self.window = WINDOW

    def test_empty_state(self):
        return
        name: str = self.window.stack.get_visible_child_name()
        self.assertEqual(name, "status")

    def test_min_stack_pages(self):
        pages_list: Gio.ListModel = self.window.stack.get_pages()
        pages: list[Adw.ViewStackPage] = [
            pages_list.get_item(i).get_name() for i in range(pages_list.get_n_items())
        ]
        expected_pages = ["trash", "status"]
        for page in expected_pages:
            self.assertTrue(page in pages)

    def test_add_toast(self):
        return
        self.window.add_toast("Test toast")

    def test_existing_actions(self):
        actions: list[str] = (
            self.window.list_actions() + self.window.get_application().list_actions()
        )
        expected_actions: list[str] = [
            "show-help-overlay",
            "preferences",
            "sync",
            "import",
            "about",
            "quit",
            "secret_notes",
        ]
        self.assertEqual(actions, expected_actions)

    def test_about_window(self):
        app: Adw.Application = self.window.get_application()
        app.activate_action("about")
        self.assertTrue(self.window.about_window.get_visible())
        self.window.about_window.close()
        self.assertFalse(self.window.about_window.get_visible())



# Copyright 2023-2024 Vlad Krupinskii <mrvladus@yandex.ru>
# SPDX-License-Identifier: MIT

from __future__ import annotations
from typing import TYPE_CHECKING

from errands.lib.gsettings import GSettings

if TYPE_CHECKING:
    from errands.widgets.task_list import TaskList
    from errands.widgets.task import Task

from datetime import datetime
from errands.lib.data import UserData
from icalendar import Calendar, Event
from errands.lib.functions import get_children
from errands.widgets.components import Box, Button, DateTime
from gi.repository import Adw, Gtk, Gdk, GObject, GtkSource, Gio  # type:ignore
from errands.lib.markup import Markup
from errands.lib.sync.sync import Sync
from errands.lib.logging import Log


class Details(Adw.Bin):
    parent: Task = None
    can_sync: bool = False

    def __init__(self, task_list: TaskList) -> None:
        super().__init__()
        self.task_list: TaskList = task_list
        self.split_view: Adw.OverlaySplitView = task_list.split_view
        self._build_ui()

    def _build_ui(self):
        # Header Bar
        hb = Adw.HeaderBar(show_title=False, show_back_button=False)
        # Back button
        back_btn = Button(
            icon_name="go-previous-symbolic",
            on_click=lambda *_: self.split_view.set_show_sidebar(False),
            visible=False,
        )
        back_btn.bind_property(
            "visible",
            self.split_view,
            "collapsed",
            GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL,
        )
        hb.pack_start(back_btn)
        # Delete button
        delete_btn = Button(
            icon_name="errands-trash-symbolic",
            on_click=self._on_delete_btn_clicked,
            tooltip_text=_("Delete"),
        )
        hb.pack_start(delete_btn)

        # Status
        self.status = Adw.StatusPage(
            icon_name="errands-info-symbolic",
            visible=True,
            vexpand=True,
            title=_("No Details"),
            description=_("Click on task to show more info"),
        )
        self.status.add_css_class("compact")
        self.status.bind_property(
            "visible",
            delete_btn,
            "visible",
            GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.INVERT_BOOLEAN,
        )

        # Colors
        colors_box = Gtk.Box(halign="center", css_classes=["toolbar"])
        colors = ["", "blue", "green", "yellow", "orange", "red", "purple", "brown"]
        for color in colors:
            btn = Gtk.Button()
            if color == "":
                btn.set_icon_name("window-close-symbolic")
                btn.set_tooltip_text(_("Clear Style"))
            else:
                btn.add_css_class("accent-color-btn")
                btn.add_css_class(f"btn-{color}")
            btn.add_css_class("circular")
            btn.connect("clicked", self._on_style_selected, color)
            colors_box.append(btn)

        # Edit group
        edit_group = Adw.PreferencesGroup(title=_("Text"))
        # Edit entry
        self.edit_entry = GtkSource.Buffer()
        Adw.StyleManager.get_default().bind_property(
            "dark",
            self.edit_entry,
            "style-scheme",
            GObject.BindingFlags.SYNC_CREATE,
            lambda _, is_dark: self.edit_entry.set_style_scheme(
                GtkSource.StyleSchemeManager.get_default().get_scheme(
                    "Adwaita-dark" if is_dark else "Adwaita"
                )
            ),
        )
        self.edit_entry.connect("changed", self._on_text_changed)
        edit_entry_text_ovrl = Gtk.Overlay(
            child=GtkSource.View(
                height_request=100,
                top_margin=12,
                bottom_margin=12,
                left_margin=12,
                right_margin=12,
                buffer=self.edit_entry,
                css_classes=["card"],
                wrap_mode=3,
            )
        )
        # Save button
        self.edit_entry_save_btn: Gtk.Button = Gtk.Button(
            icon_name="errands-select-symbolic",
            css_classes=["circular", "suggested-action"],
            halign=Gtk.Align.END,
            valign=Gtk.Align.END,
            margin_bottom=6,
            margin_end=6,
            tooltip_text=_("Save"),
            visible=False,
        )
        self.edit_entry_save_btn.connect("clicked", lambda *_: self.save())
        edit_entry_text_ovrl.add_overlay(self.edit_entry_save_btn)
        edit_group.add(edit_entry_text_ovrl)
        # Copy button
        text_copy_btn = Gtk.Button(
            icon_name="errands-copy-symbolic",
            valign="center",
            tooltip_text=_("Copy Text"),
            css_classes=["flat"],
        )
        text_copy_btn.connect("clicked", self._on_copy_clicked, self.edit_entry)
        edit_group.set_header_suffix(text_copy_btn)

        # Notes group
        notes_group = Adw.PreferencesGroup(title=_("Notes"))
        # Notes entry
        notes_source_view = GtkSource.View(
            height_request=200,
            top_margin=12,
            bottom_margin=12,
            left_margin=12,
            right_margin=12,
            wrap_mode=3,
            css_classes=["card"],
        )
        self.notes = notes_source_view.get_buffer()
        Adw.StyleManager.get_default().bind_property(
            "dark",
            self.notes,
            "style-scheme",
            GObject.BindingFlags.SYNC_CREATE,
            lambda _, is_dark: self.notes.set_style_scheme(
                GtkSource.StyleSchemeManager.get_default().get_scheme(
                    "Adwaita-dark" if is_dark else "Adwaita"
                )
            ),
        )
        lm: GtkSource.LanguageManager = GtkSource.LanguageManager.get_default()
        self.notes.set_language(lm.get_language("markdown"))
        self.notes.connect("changed", self._on_notes_changed)
        notes_ovrl = Gtk.Overlay(child=notes_source_view)
        # Save button
        self.notes_save_btn: Gtk.Button = Gtk.Button(
            icon_name="errands-select-symbolic",
            css_classes=["circular", "suggested-action"],
            halign=Gtk.Align.END,
            valign=Gtk.Align.END,
            margin_bottom=6,
            margin_end=6,
            tooltip_text=_("Save"),
            visible=False,
        )
        self.notes_save_btn.connect("clicked", lambda *_: self.save())
        notes_ovrl.add_overlay(self.notes_save_btn)
        # Copy button
        notes_copy_btn = Gtk.Button(
            icon_name="errands-copy-symbolic",
            valign="center",
            tooltip_text=_("Copy Text"),
            css_classes=["flat"],
        )
        notes_copy_btn.connect("clicked", self._on_copy_clicked, self.notes)
        notes_group.set_header_suffix(notes_copy_btn)
        notes_group.add(notes_ovrl)

        # Properties group
        props_group = Adw.PreferencesGroup(title=_("Properties"))

        # Start date row
        self.start_datetime_row = Adw.ActionRow(title=_("Not Set"), subtitle=_("Start"))
        self.start_datetime = DateTime()
        self.start_datetime.connect("changed", self._on_start_time_changed)
        self.start_datetime_row.add_suffix(
            Gtk.MenuButton(
                valign="center",
                icon_name="errands-calendar-symbolic",
                tooltip_text=_("Set Date"),
                popover=Gtk.Popover(
                    child=Gtk.ScrolledWindow(
                        propagate_natural_height=True,
                        propagate_natural_width=True,
                        child=self.start_datetime,
                    )
                ),
                css_classes=["flat"],
            )
        )
        props_group.add(self.start_datetime_row)

        # End date row
        self.end_datetime_row = Adw.ActionRow(title=_("Not Set"), subtitle=_("Due"))
        self.end_datetime = DateTime()
        self.end_datetime.connect("changed", self._on_end_time_changed)
        self.end_datetime_row.add_suffix(
            Gtk.MenuButton(
                valign="center",
                icon_name="errands-calendar-symbolic",
                tooltip_text=_("Set Date"),
                popover=Gtk.Popover(child=self.end_datetime),
                css_classes=["flat"],
            )
        )
        props_group.add(self.end_datetime_row)

        # Complete % row
        percent_complete_row = Adw.ActionRow(title=_("Complete %"))
        self.percent_complete = Gtk.SpinButton(
            valign="center",
            adjustment=Gtk.Adjustment(lower=0, upper=100, step_increment=1),
        )
        self.percent_complete.connect("changed", self._on_percent_complete_changed)
        percent_complete_row.add_suffix(self.percent_complete)
        props_group.add(percent_complete_row)

        # Priority row
        priority_row = Adw.ActionRow(title=_("Priority"))
        self.priority = Gtk.SpinButton(
            valign="center",
            adjustment=Gtk.Adjustment(lower=0, upper=9, step_increment=1),
        )
        self.priority.connect("changed", self._on_priority_changed)
        priority_row.add_suffix(self.priority)
        props_group.add(priority_row)

        # Tags group
        self.tags = Adw.PreferencesGroup(title=_("Tags"))
        # Tags entry
        self.tag_entry = Adw.EntryRow(title=_("Add Tag"))
        self.tag_entry.connect("entry-activated", self._on_tag_added)
        self.tags.add(self.tag_entry)

        # Export group
        misc_group = Adw.PreferencesGroup(title=_("Export"))
        open_cal_btn = Button(
            icon_name="errands-share-symbolic",
            on_click=self._on_export,
            valign="center",
            css_classes=["flat"],
        )
        open_cal_row = Adw.ActionRow(
            title=_("Export"),
            subtitle=_("Save Task as .ics file"),
            activatable_widget=open_cal_btn,
        )
        open_cal_row.add_suffix(open_cal_btn)
        misc_group.add(open_cal_row)

        # Groups box
        p_box = Box(
            children=[
                colors_box,
                edit_group,
                notes_group,
                props_group,
                self.tags,
                misc_group,
            ],
            orientation="vertical",
            spacing=12,
            visible=False,
        )
        p_box.bind_property(
            "visible",
            self.status,
            "visible",
            GObject.BindingFlags.SYNC_CREATE
            | GObject.BindingFlags.INVERT_BOOLEAN
            | GObject.BindingFlags.BIDIRECTIONAL,
        )
        # Toolbar View
        toolbar_view = Adw.ToolbarView(
            content=Gtk.ScrolledWindow(
                child=Adw.Clamp(
                    child=Box(children=[self.status, p_box], orientation="vertical"),
                    margin_start=12,
                    margin_end=12,
                    margin_top=6,
                    margin_bottom=24,
                )
            )
        )
        toolbar_view.add_top_bar(hb)
        self.set_child(toolbar_view)

    def add_tag(self, text: str) -> None:
        if text == "":
            return
        tag = Adw.ActionRow(title=text)
        delete_btn = Gtk.Button(
            icon_name="window-close-symbolic",
            valign="center",
            css_classes=["flat", "circular"],
        )
        delete_btn.connect("clicked", self._on_tag_deleted, tag)
        tag.add_suffix(delete_btn)
        self.tags.add(tag)

    def save(self):
        Log.debug("Details: Save")
        self.edit_entry_save_btn.set_visible(False)
        self.notes_save_btn.set_visible(False)

        # Text
        old_text: str = self.parent.get_prop("text")
        text: str = self.edit_entry.props.text
        if text.strip(" \n\t") and text != old_text:
            self.parent.task_row.task_row.set_title(
                Markup.find_url(Markup.escape(text))
            )

        # Save props
        self.parent.update_props(
            [
                "text",
                "notes",
                "completed",
                "start_date",
                "end_date",
                "percent_complete",
                "priority",
                "synced",
            ],
            [
                text if text.strip(" \n\t") and text != old_text else old_text,
                self.notes.props.text,
                self.parent.task_row.complete_btn.get_active(),
                self.start_datetime.get_datetime(),
                self.end_datetime.get_datetime(),
                self.percent_complete.get_value_as_int(),
                self.priority.get_value_as_int(),
                False,
            ],
        )

        self.parent.update_ui()
        Sync.sync(False)

    def update_info(self, parent: Task):
        # Show status on empty task
        if parent == None:
            self.status.set_visible(True)
            return

        self.can_sync = False
        self.parent = parent

        Log.debug("Details: Update info")

        # Edit text
        self.edit_entry.set_text(self.parent.get_prop("text"))
        # Notes
        self.notes.set_text(self.parent.get_prop("notes"))
        # Datetime
        self.start_datetime.set_datetime(self.parent.get_prop("start_date"))
        self.start_datetime_row.set_title(self.start_datetime.get_human_datetime())
        self.end_datetime.set_datetime(self.parent.get_prop("end_date"))
        self.end_datetime_row.set_title(self.end_datetime.get_human_datetime())
        # Percent complete
        self.percent_complete.set_value(self.parent.get_prop("percent_complete"))
        # Priority
        self.priority.set_value(self.parent.get_prop("priority"))
        self.status.set_visible(False)
        # Tags
        # Remove old
        for i, tag in enumerate(get_children(self.tag_entry.get_parent())):
            # Skip entry
            if i == 0:
                continue
            self.tags.remove(tag)
        # Add new
        for tag in self.parent.get_prop("tags").split(","):
            self.add_tag(tag)
        self.can_sync = True

    def _on_text_changed(self, buffer: Gtk.TextBuffer):
        if not self.can_sync:
            return
        if buffer.props.text != self.parent.get_prop("text"):
            if GSettings.get("sync-provider") != 0:
                self.edit_entry_save_btn.set_visible(True)
            else:
                self.save()

    def _on_notes_changed(self, buffer: GtkSource.Buffer):
        if not self.can_sync:
            return
        if buffer.props.text != self.parent.get_prop("notes"):
            if GSettings.get("sync-provider") != 0:
                self.notes_save_btn.set_visible(True)
            else:
                self.save()

    def _on_percent_complete_changed(self, _):
        if not self.can_sync:
            return
        self.save()

    def _on_priority_changed(self, row: Adw.SpinRow):
        if not self.can_sync:
            return
        self.save()

    def _on_start_time_changed(self, *args):
        if not self.can_sync:
            return

        Log.debug("Details: change start time")

        sdt: int = self.start_datetime.get_datetime_as_int()
        edt: int = self.end_datetime.get_datetime_as_int()
        if sdt > edt and edt != 0:
            self.end_datetime.set_datetime(self.start_datetime.get_datetime())
            self.end_datetime_row.set_title(self.end_datetime.get_human_datetime())
        self.start_datetime_row.set_title(self.start_datetime.get_human_datetime())
        self.save()

    def _on_end_time_changed(self, *args):
        if not self.can_sync:
            return

        Log.debug("Details: change end time")

        sdt: int = self.start_datetime.get_datetime_as_int()
        edt: int = self.end_datetime.get_datetime_as_int()
        if edt < sdt and sdt != 0 and edt != 0:
            self.start_datetime.set_datetime(self.end_datetime.get_datetime())
            self.start_datetime_row.set_title(self.start_datetime.get_human_datetime())
        self.end_datetime_row.set_title(self.end_datetime.get_human_datetime())
        self.save()

    def _on_copy_clicked(self, _btn, buffer: Gtk.TextBuffer):
        Log.info("Details: Copy to clipboard")
        Gdk.Display.get_default().get_clipboard().set(buffer.props.text)
        self.task_list.window.add_toast(_("Copied to Clipboard"))

    def _on_delete_btn_clicked(self, _btn):
        self.parent.delete()
        self.status.set_visible(True)

    def _on_export(self, _btn):
        def _confirm(dialog, res):
            try:
                file = dialog.save_finish(res)
            except:
                Log.debug("List: Export cancelled")
                return

            Log.info(f"Task: Export '{self.parent.uid}'")

            task = [
                i
                for i in UserData.get_tasks_as_dicts(self.parent.list_uid)
                if i["uid"] == self.parent.uid
            ][0]
            calendar = Calendar()
            event = Event()
            event.add("uid", task["uid"])
            event.add("summary", task["text"])
            if task["notes"]:
                event.add("description", task["notes"])
            event.add("priority", task["priority"])
            if task["tags"]:
                event.add("categories", task["tags"])
            event.add("percent-complete", task["percent_complete"])
            if task["color"]:
                event.add("x-errands-color", task["color"])
            event.add(
                "dtstart",
                (
                    datetime.fromisoformat(task["start_date"])
                    if task["start_date"]
                    else datetime.now()
                ),
            )
            if task["end_date"]:
                event.add("dtend", datetime.fromisoformat(task["end_date"]))
            calendar.add_component(event)

            with open(file.get_path(), "wb") as f:
                f.write(calendar.to_ical())
            self.window.add_toast(_("Exported"))

        dialog = Gtk.FileDialog(initial_name=f"{self.parent.uid}.ics")
        dialog.save(self.task_list.window, None, _confirm)

    def _on_style_selected(self, btn: Gtk.Button, color: str) -> None:
        """
        Apply accent color
        """
        for c in self.parent.main_box.get_css_classes():
            if "task-" in c:
                self.parent.main_box.remove_css_class(c)
                break
        if color != "":
            self.parent.main_box.add_css_class(f"task-{color}")
        self.parent.update_props(["color", "synced"], [color, False])
        Sync.sync()

    # --- Tags --- #

    def _on_tag_added(self, entry: Adw.EntryRow) -> None:
        text = entry.get_text().strip(" \n\t")
        if text == "":
            return
        Log.debug("Details: Add tag")
        self.add_tag(entry.get_text())
        entry.set_text("")
        self._save_tags()

    def _on_tag_deleted(self, btn, tag):
        Log.debug("Details: Remove tag")
        self.tags.remove(tag)
        self._save_tags()

    def _save_tags(self):
        self.parent.update_props(
            ["tags", "synced"],
            [
                ",".join(
                    [
                        r.get_title()
                        for r in get_children(self.tag_entry.get_parent())[1:]
                    ]
                ),
                False,
            ],
        )
        Sync.sync()



# Copyright 2020 gdavid7 https://github.com/gdavid7/cryptocode
# Copyright 2024 Vlad Krupinskii <mrvladus@yandex.ru>
# SPDX-License-Identifier: MIT

from base64 import b64encode, b64decode
import hashlib
from Cryptodome.Cipher import AES
from Cryptodome.Random import get_random_bytes


def encrypt(message: str, password: str) -> str:
    """Encrypt string with password"""

    # generate a random salt
    salt: bytes = get_random_bytes(AES.block_size)

    # use the Scrypt KDF to get a private key from the password
    private_key: bytes = hashlib.scrypt(
        password.encode(),
        salt=salt,
        n=2**14,
        r=8,
        p=1,
        dklen=32,
    )

    # create cipher config
    cipher_config = AES.new(private_key, AES.MODE_GCM)

    # return a dictionary with the encrypted text
    (cipher_text, tag) = cipher_config.encrypt_and_digest(bytes(message, "utf-8"))
    encryptedDict: dict[str, str] = {
        "cipher_text": b64encode(cipher_text).decode("utf-8"),
        "salt": b64encode(salt).decode("utf-8"),
        "nonce": b64encode(cipher_config.nonce).decode("utf-8"),
        "tag": b64encode(tag).decode("utf-8"),
    }
    encryptedString: str = (
        encryptedDict["cipher_text"]
        + "*"
        + encryptedDict["salt"]
        + "*"
        + encryptedDict["nonce"]
        + "*"
        + encryptedDict["tag"]
    )
    return encryptedString


def decrypt(enc_message: str, password: str) -> str | None:
    """Decrypt string with password"""

    enc_dict: list[str] = enc_message.split("*")
    try:
        enc_dict: dict[str, str] = {
            "cipher_text": enc_dict[0],
            "salt": enc_dict[1],
            "nonce": enc_dict[2],
            "tag": enc_dict[3],
        }

        # decode the dictionary entries from base64
        salt: bytes = b64decode(enc_dict["salt"])
        cipher_text: bytes = b64decode(enc_dict["cipher_text"])
        nonce: bytes = b64decode(enc_dict["nonce"])
        tag: bytes = b64decode(enc_dict["tag"])

        # generate the private key from the password and salt
        private_key: bytes = hashlib.scrypt(
            password.encode(),
            salt=salt,
            n=2**14,
            r=8,
            p=1,
            dklen=32,
        )

        # create the cipher config
        cipher = AES.new(private_key, AES.MODE_GCM, nonce=nonce)

        # decrypt the cipher text
        decrypted: bytes = cipher.decrypt_and_verify(cipher_text, tag)
    except:
        return None

    return decrypted.decode("UTF-8")



import urllib.request
import json
import uuid


class SyncProviderTodoist:
    def __init__(self, api_token):
        self.api_token = api_token
        self.base_url = "https://api.todoist.com/sync/v9/sync"

    def sync(self, commands):
        url = self.base_url
        headers = {
            "Authorization": f"Bearer {self.api_token}",
            "Content-Type": "application/json",
        }
        data = {"token": self.api_token, "commands": json.dumps(commands)}
        req = urllib.request.Request(
            url, headers=headers, data=json.dumps(data).encode("utf-8")
        )
        with urllib.request.urlopen(req) as response:
            response_json = json.loads(response.read().decode("utf-8"))
        return response_json

    def get_projects(self):
        commands = [{"type": "projects/get_data"}]
        response = self.sync(commands)
        projects = response["Projects"]
        return projects

    def create_task(self, project_id, content):
        commands = [
            {
                "type": "item_add",
                "temp_id": self._generate_temp_id(),
                "uuid": self._generate_uuid(),
                "args": {"content": content, "project_id": project_id},
            }
        ]
        response = self.sync(commands)
        task = response["TempIdMapping"][0]
        return task

    def complete_task(self, task_id):
        commands = [
            {
                "type": "item_complete",
                "uuid": self._generate_uuid(),
                "args": {"id": task_id},
            }
        ]
        response = self.sync(commands)
        return response["Items"][0]["id"] == task_id

    def _generate_temp_id(self):
        return f"temp_id_{uuid.uuid4().hex}"

    def _generate_uuid(self):
        return str(uuid.uuid4())



# Copyright 2023 Vlad Krupinskii <mrvladus@yandex.ru>
# SPDX-License-Identifier: MIT

from gi.repository import Adw, Gtk  # type:ignore


def property(
    obj: Gtk.Widget,
    prop: str,
    val_from,
    val_to,
    time_ms: int,
) -> None:
    """Animate widget property"""

    def callback(value, _) -> None:
        obj.set_property(prop, value)

    animation = Adw.TimedAnimation.new(
        obj,
        val_from,
        val_to,
        time_ms,
        Adw.CallbackAnimationTarget.new(callback, None),
    )
    animation.play()


def scroll(win: Gtk.ScrolledWindow, scroll_down: bool = True, widget=None) -> None:
    """Animate scrolling"""

    adj = win.get_vadjustment()

    def callback(value, _) -> None:
        adj.set_property("value", value)

    if not widget:
        # Scroll to the top or bottom
        scroll_to = adj.get_upper() if scroll_down else adj.get_lower()
    else:
        scroll_to = widget.get_allocation().height + adj.get_value()

    animation = Adw.TimedAnimation.new(
        win,
        adj.get_value(),
        scroll_to,
        250,
        Adw.CallbackAnimationTarget.new(callback, None),
    )
    animation.play()



# Copyright 2023-2024 Vlad Krupinskii <mrvladus@yandex.ru>
# SPDX-License-Identifier: MIT

from typing import TypedDict
from errands.lib.logging import Log


class GoaCredentials(TypedDict):
    url: str
    username: str
    password: str


def get_goa_credentials(acc_name: str) -> GoaCredentials | None:
    """
    If Gnome Online Accounts is installed, try to get account info.
    Only for password based account yet.
    """

    try:
        import gi  # type:ignore

        gi.require_version("Goa", "1.0")
        from gi.repository import Goa  # type:ignore
    except (ValueError, ModuleNotFoundError):
        Log.debug("Gnome Online Accounts is not installed. Skipping...")
        return None

    try:
        client: Goa.Client = Goa.Client.new_sync(None)
        accounts: list[Goa.ObjectProxy] = client.get_accounts()
        for account in accounts:
            acc: Goa.AccountProxy = account.get_account()
            name: str = acc.get_cached_property("ProviderName").get_string()
            if name == acc_name:
                Log.debug(f"GOA: Getting data for {acc_name}")
                username: str = (
                    acc.get_cached_property("PresentationIdentity")
                    .get_string()
                    .split("@")[:-1]
                )
                username = "".join(username)
                password = account.get_password_based().call_get_password_sync(
                    arg_id=acc.get_cached_property("Id").get_string()
                )
                # Get url
                try:
                    url: str = (
                        account.get_calendar()
                        .get_cached_property("Uri")
                        .get_string()
                        .replace(username + "@", "")
                    )
                except:
                    url: str = (
                        "https://"
                        + (
                            acc.get_cached_property("PresentationIdentity")
                            .get_string()
                            .split("@")[-1]
                        )
                        + "/remote.php/dav/"
                    )

                return GoaCredentials(url=url, username=username, password=password)
    except Exception as e:
        Log.error(f"GOA: Can't get credentials. {e}")
        return None

    return None



# Copyright 2023 Vlad Krupinskii <mrvladus@yandex.ru>
# SPDX-License-Identifier: MIT

import os
from gi.repository import GLib  # type:ignore
from __main__ import VERSION


class Log:
    """Logging class"""

    data_dir: str = os.path.join(GLib.get_user_data_dir(), "errands")
    log_file: str = os.path.join(data_dir, "log.txt")
    log_old_file: str = os.path.join(data_dir, "log.old.txt")

    @classmethod
    def init(self):
        # Create data dir
        if not os.path.exists(self.data_dir):
            os.mkdir(self.data_dir)
        # Copy old log
        if os.path.exists(self.log_file):
            os.rename(self.log_file, self.log_old_file)
        # Start new log
        self.debug("Starting Errands " + VERSION)

    @classmethod
    def debug(self, msg: str) -> None:
        print(f"\033[33;1m[DEBUG]\033[0m {msg}")
        self._log(self, f"[DEBUG] {msg}")

    @classmethod
    def error(self, msg: str) -> None:
        print(f"\033[31;1m[ERROR]\033[0m {msg}")
        self._log(self, f"[ERROR] {msg}")

    @classmethod
    def info(self, msg: str) -> None:
        print(f"\033[32;1m[INFO]\033[0m {msg}")
        self._log(self, f"[INFO] {msg}")

    def _log(self, msg: str) -> None:
        try:
            with open(self.log_file, "a") as f:
                f.write(msg + "\n")
        except OSError:
            self.error("Can't write to the log file")



#!@PYTHON@

# Copyright 2023 Vlad Krupinskii <mrvladus@yandex.ru>
# SPDX-License-Identifier: MIT

import os
import sys
import signal
import locale
import gettext
import gi  # type:ignore

gi.require_version("Gtk", "4.0")
gi.require_version("Adw", "1")
gi.require_version("Secret", "1")
gi.require_version("GtkSource", "5")

from gi.repository import Adw, Gio  # type:ignore

APP_ID = "@APP_ID@"
VERSION = "@VERSION@"
PREFIX = "@PREFIX@"
PROFILE = "@PROFILE@"
pkgdatadir = "@pkgdatadir@"
localedir = "@localedir@"

sys.path.insert(1, pkgdatadir)
signal.signal(signal.SIGINT, signal.SIG_DFL)
gettext.install("errands", localedir)
locale.bindtextdomain("errands", localedir)
locale.textdomain("errands")


def main() -> None:
    resource = Gio.Resource.load(os.path.join(pkgdatadir, "errands.gresource"))
    resource._register()

    from errands.lib.logging import Log
    from errands.lib.data import UserData

    Log.init()
    UserData.init()

    sys.exit(Application().run(sys.argv))


class Application(Adw.Application):
    def __init__(self) -> None:
        super().__init__(
            application_id=APP_ID,
            flags=Gio.ApplicationFlags.DEFAULT_FLAGS,
        )
        self.set_resource_base_path("/io/github/mrvladus/Errands/")

    def do_activate(self) -> None:
        from errands.widgets.window import Window

        Window(application=self)

        # Run tests
        if PROFILE == "development":
            from errands.tests.tests import run_tests

            run_tests()


if __name__ == "__main__":
    main()



# Copyright 2023 Vlad Krupinskii <mrvladus@yandex.ru>
# SPDX-License-Identifier: MIT

from typing import Callable
from threading import Thread
from gi.repository import Gtk  # type:ignore


def get_children(obj: Gtk.Widget) -> list[Gtk.Widget]:
    """
    Get list of widget's children
    """

    children: list[Gtk.Widget] = []
    child: Gtk.Widget = obj.get_first_child()
    while child:
        children.append(child)
        child = child.get_next_sibling()
    return children


def threaded(function: Callable):
    """
    Decorator to run function in thread.
    Use GLib.idle_add(func) as callback at the end of threaded function
    if you need to change UI from thread.
    It's needed to be called to make changes in UI thread safe.
    """

    def wrapper(*args, **kwargs):
        Thread(target=function, args=args, kwargs=kwargs, daemon=True).start()

    return wrapper



# Copyright 2023 Vlad Krupinskii <mrvladus@yandex.ru>
# SPDX-License-Identifier: MIT

from __future__ import annotations
from typing import TYPE_CHECKING

from errands.widgets.task_list import TaskList

if TYPE_CHECKING:
    from errands.widgets.window import Window

from errands.lib.data import UserData
from errands.lib.functions import get_children
from errands.widgets.components import Box, Button
from gi.repository import Adw, Gtk, GObject  # type:ignore
from errands.widgets.task import Task
from errands.lib.sync.sync import Sync
from errands.lib.logging import Log


class Trash(Adw.Bin):
    def __init__(self, window: Window):
        super().__init__()
        self.window: Window = window
        self.stack: Adw.ViewStack = window.stack
        self._build_ui()
        self.update_status()

    def _build_ui(self) -> None:
        # Headerbar
        hb: Adw.HeaderBar = Adw.HeaderBar(
            title_widget=Adw.WindowTitle(title=_("Trash"))
        )
        # Clear button
        clear_btn = Button(
            icon_name="errands-trash-symbolic",
            on_click=self.on_trash_clear,
            valign="center",
            tooltip_text=_("Clear"),
        )
        hb.pack_start(clear_btn)
        # Restore button
        restore_btn = Button(
            icon_name="edit-redo-symbolic",
            on_click=self.on_trash_restore,
            valign="center",
            tooltip_text=_("Restore All"),
        )
        hb.pack_end(restore_btn)
        # Status
        self.status = Adw.StatusPage(
            icon_name="errands-trash-symbolic",
            title=_("Empty Trash"),
            description=_("No deleted items"),
            vexpand=True,
        )
        self.status.add_css_class("compact")
        # Trash list
        self.trash_list = Gtk.Box(
            orientation="vertical",
            margin_top=12,
            margin_bottom=12,
            margin_start=12,
            margin_end=12,
            spacing=12,
            css_classes=["boxed-list"],
            vexpand=False,
        )
        scrl = Gtk.ScrolledWindow(
            child=Adw.Clamp(
                child=self.trash_list,
                maximum_size=850,
                tightening_threshold=300,
            ),
            visible=False,
            propagate_natural_height=True,
        )
        scrl.bind_property(
            "visible",
            self.status,
            "visible",
            GObject.BindingFlags.INVERT_BOOLEAN | GObject.BindingFlags.BIDIRECTIONAL,
        )
        scrl.bind_property(
            "visible", clear_btn, "visible", GObject.BindingFlags.SYNC_CREATE
        )
        scrl.bind_property(
            "visible", restore_btn, "visible", GObject.BindingFlags.SYNC_CREATE
        )
        # Toolbar View
        toolbar_view = Adw.ToolbarView(
            content=Box(children=[self.status, scrl], orientation="vertical")
        )
        toolbar_view.add_top_bar(hb)
        self.set_child(toolbar_view)

    def trash_add(self, task_widget) -> None:
        self.trash_list.append(TrashItem(task_widget, self))
        self.status.set_visible(False)

    def update_status(self) -> None:
        self.status.set_visible(len([t for t in get_children(self.trash_list)]) == 0)

    def on_trash_clear(self, btn) -> None:
        def _confirm(_, res) -> None:
            if res == "cancel":
                Log.debug("Clear Trash cancelled")
                return

            Log.info("Trash: Clear")

            UserData.run_sql(
                f"""UPDATE tasks
                SET deleted = 1
                WHERE trash = 1""",
            )

            # Remove tasks
            for row in get_children(self.trash_list):
                row.task_widget.purge()
                self.trash_list.remove(row)

            self.status.set_visible(True)
            self.update_status()
            # Sync
            Sync.sync()

        Log.debug("Trash: Show confirm dialog")

        dialog = Adw.MessageDialog(
            transient_for=self.window,
            hide_on_close=True,
            heading=_("Are you sure?"),
            body=_("Tasks will be permanently deleted"),
            default_response="delete",
            close_response="cancel",
        )
        dialog.add_response("cancel", _("Cancel"))
        dialog.add_response("delete", _("Delete"))
        dialog.set_response_appearance("delete", Adw.ResponseAppearance.DESTRUCTIVE)
        dialog.connect("response", _confirm)
        dialog.show()

    def on_trash_restore(self, _) -> None:
        """
        Remove trash items and restore all tasks
        """

        Log.info("Trash: Restore")

        # Get all tasks
        tasks: list[Task] = []
        task_lists: list[TaskList] = []
        pages = self.stack.get_pages()
        for i in range(pages.get_n_items()):
            child = pages.get_item(i).get_child()
            if hasattr(child, "get_all_tasks"):
                task_lists.append(child)
                tasks.extend(child.get_all_tasks())

        for task in tasks:
            task.update_props(["trash"], [False])

        self.window.sidebar.task_lists.update_ui()

        for row in get_children(self.trash_list):
            self.trash_list.remove(row)

        self.update_status()


class TrashItem(Adw.Bin):
    def __init__(self, task_widget: Task, trash: Trash) -> None:
        super().__init__()
        self.task_widget: Task = task_widget
        self.uid: str = task_widget.uid
        self.trash: Trash = trash
        self.trash_list: Gtk.Box = trash.trash_list
        self._build_ui()

    def _build_ui(self) -> None:
        self.add_css_class("card")
        row: Adw.ActionRow = Adw.ActionRow(
            title=self.task_widget.get_prop("text"),
            css_classes=["rounded-corners"],
            height_request=60,
            title_selectable=True,
        )
        self.task_widget.task_row.connect(
            "notify::title", lambda *_: row.set_title(self.task_widget.get_prop("text"))
        )
        self.task_widget.task_list.title.bind_property(
            "title",
            row,
            "subtitle",
            GObject.BindingFlags.SYNC_CREATE,
        )
        row.add_suffix(
            Button(
                icon_name="emblem-ok-symbolic",
                on_click=self.on_restore,
                tooltip_text=_("Restore"),
                valign="center",
                css_classes=["flat", "circular"],
            )
        )
        box: Gtk.ListBox = Gtk.ListBox(
            selection_mode=0,
            css_classes=["rounded-corners"],
            accessible_role=Gtk.AccessibleRole.PRESENTATION,
        )
        box.append(row)
        self.set_child(box)

    def on_restore(self, _) -> None:
        """Restore task and its parents"""

        Log.info(f"Restore task: {self.uid}")

        tasks: list[Task] = []
        tasks.append(self.task_widget)
        tasks.extend(self.task_widget.get_parents_tree())
        for task in tasks:
            task.update_props(["trash"], [False])

        tasks_uids: list[str] = [t.uid for t in tasks]

        for item in get_children(self.trash_list):
            if item.uid in tasks_uids:
                self.trash_list.remove(item)

        self.task_widget.task_list.update_ui()
        self.trash.update_status()



# Copyright 2023-2024 Vlad Krupinskii <mrvladus@yandex.ru>
# SPDX-License-Identifier: MIT

from __future__ import annotations
from typing import TYPE_CHECKING, TypedDict

from errands.lib.functions import get_children

if TYPE_CHECKING:
    from errands.widgets.window import Window

from uuid import uuid4
from errands.lib.gsettings import GSettings
from errands.lib.data import UserData
from errands.widgets.components import Box
from errands.lib.encryption import encrypt, decrypt
from gi.repository import Adw, Gtk, GtkSource, GObject  # type:ignore


class NoteDict(TypedDict):
    text: str
    uid: str


class SecretNote(Gtk.ListBoxRow):
    def __init__(self, notes_page: SecretNotesPage, note: NoteDict):
        super().__init__()
        self.notes_page: SecretNotesPage = notes_page
        self.notes_list: Gtk.ListBox = notes_page.notes_list
        self.password: str = notes_page.password
        self.uid: str = note["uid"]
        self.text: str = note["text"]
        self._build_ui()

    def _build_ui(self):
        self.set_activatable(False)
        self.set_margin_top(6)
        self.set_margin_bottom(6)
        self.set_margin_start(12)
        self.set_margin_end(12)

        # Text view
        self.text_view = GtkSource.View(
            css_classes=["card", "secret-note"],
            height_request=100,
            top_margin=12,
            bottom_margin=12,
            right_margin=12,
            left_margin=12,
            wrap_mode=Gtk.WrapMode.WORD_CHAR,
        )
        self.buffer = self.text_view.get_buffer()
        self.buffer.props.text = self.text
        lm: GtkSource.LanguageManager = GtkSource.LanguageManager.get_default()
        self.buffer.set_language(lm.get_language("markdown"))
        self.buffer.connect("changed", self._on_text_changed)

        # Delete button
        delete_btn = Gtk.Button(
            icon_name="errands-trash-symbolic",
            css_classes=["flat", "circular", "error"],
            halign=Gtk.Align.END,
            valign=Gtk.Align.END,
            margin_end=6,
            margin_bottom=6,
        )
        delete_btn.connect("clicked", self._on_delete_clicked)

        # Overlay
        overlay = Gtk.Overlay(child=self.text_view)
        overlay.add_overlay(delete_btn)

        self.set_child(overlay)

    def _on_text_changed(self, buffer: GtkSource.Buffer):
        text: str = encrypt(buffer.props.text, self.password)
        UserData.execute(
            "UPDATE secret_notes SET text = ? WHERE uid = ?", (text, self.uid)
        )

    def _on_delete_clicked(self, _btn):
        UserData.execute("DELETE FROM secret_notes WHERE uid = ?", (self.uid,))
        self.notes_list.remove(self)
        self.notes_page.update_ui()


class SecretNotesPage(Adw.Bin):
    password: str = None

    def __init__(self, password: str):
        super().__init__()
        self.password: str = password
        self._build_ui()
        self._load_notes()
        self.update_ui()

    def _add_note(self, note: NoteDict):
        self.notes_list.append(SecretNote(self, note))

    def _build_ui(self):
        # Headerbar
        hb: Adw.HeaderBar = Adw.HeaderBar(
            title_widget=Adw.WindowTitle(title=_("Secret Notes"))
        )

        # Add note button
        self.add_btn: Gtk.Button = Gtk.Button(
            child=Adw.ButtonContent(icon_name="errands-add-symbolic", label=_("Add"))
        )
        self.add_btn.connect("clicked", self._on_add_note_clicked)
        hb.pack_start(self.add_btn)

        self.notes_list: Gtk.ListBox = Gtk.ListBox(
            selection_mode=0, css_classes=["background"]
        )

        # Status
        self.status_page = Adw.StatusPage(
            title="Add Note",
            icon_name="errands-notes-symbolic",
            vexpand=True,
        )

        # Scrolled window
        self.scrl = Gtk.ScrolledWindow(
            child=Adw.Clamp(
                child=self.notes_list,
                margin_bottom=36,
                maximum_size=1000,
                tightening_threshold=300,
            ),
            propagate_natural_height=True,
        )
        self.scrl.bind_property(
            "visible",
            self.status_page,
            "visible",
            GObject.BindingFlags.SYNC_CREATE
            | GObject.BindingFlags.BIDIRECTIONAL
            | GObject.BindingFlags.INVERT_BOOLEAN,
        )

        # Toolbar view
        toolbar_view: Adw.ToolbarView = Adw.ToolbarView(
            content=Box(
                children=[
                    self.status_page,
                    self.scrl,
                ],
                orientation=Gtk.Orientation.VERTICAL,
            )
        )
        toolbar_view.add_top_bar(hb)
        self.set_child(toolbar_view)

    def _load_notes(self):
        notes: list[NoteDict] = [
            NoteDict(text=decrypt(i[1], self.password), uid=i[0])
            for i in UserData.execute("SELECT uid, text FROM secret_notes", fetch=True)
        ]
        for note in notes:
            self._add_note(note)

    def _on_add_note_clicked(self, _btn):
        uid: str = str(uuid4())
        UserData.execute(
            "INSERT INTO secret_notes (text, uid) VALUES (?, ?)", ("", uid)
        )
        self._add_note(NoteDict(text="", uid=uid))
        self.update_ui()

    def update_ui(self):
        self.status_page.set_visible(not get_children(self.notes_list))


class SecretNotesPasswordPage(Adw.Bin):
    attempts: int = 0

    def __init__(self, notes_win: SecretNotesWindow):
        super().__init__()
        self.notes_win: SecretNotesWindow = notes_win
        self._build_ui()

    def _build_ui(self):
        password: str | None = GSettings.get_secret("errands_secret_notes")

        # Status page
        self.status_page = Adw.StatusPage(
            title=_("Secret Notes"),
            description=(
                _("Enter password") if password != None else _("Create password")
            ),
            icon_name="errands-private-symbolic",
            height_request=300,
        )

        # Entry
        password_entry = Adw.PasswordEntryRow(
            title=_("Password"),
            css_classes=["card"],
            height_request=60,
            margin_end=12,
            margin_start=12,
        )
        password_entry.connect("entry-activated", self._on_password_entered)

        # Attempts bar
        self.attempts_bar = Gtk.ProgressBar(
            text=_("Attempts"),
            show_text=True,
            margin_top=12,
            margin_end=12,
            margin_start=12,
        )
        self.attempts_bar.add_css_class("secret-notes-progressbar")
        self.attempts_bar.set_fraction(0.5)
        self.attempts_rev = Gtk.Revealer(child=self.attempts_bar)

        # Delete notes button
        self.delete_btn = Gtk.Button(
            label=_("Delete Notes"),
            css_classes=["pill", "destructive-action"],
            halign=Gtk.Align.CENTER,
            margin_top=24,
            visible=password != None,
        )
        self.delete_btn.connect("clicked", self._on_delete_clicked)

        # Toolbar view
        toolbar_view: Adw.ToolbarView = Adw.ToolbarView(
            content=Adw.Clamp(
                child=Box(
                    children=[
                        self.status_page,
                        password_entry,
                        self.attempts_rev,
                        self.delete_btn,
                    ],
                    orientation=Gtk.Orientation.VERTICAL,
                ),
                margin_bottom=36,
                valign=Gtk.Align.CENTER,
            )
        )
        toolbar_view.add_top_bar(Adw.HeaderBar(show_title=False))
        self.set_child(toolbar_view)

    def _on_password_entered(self, entry: Adw.PasswordEntryRow):
        password: str = entry.get_text()

        # Ignore empty password
        if not password:
            return

        # Get user password
        notes_password: str | None = GSettings.get_secret("errands_secret_notes")

        # Create password if not exists and open notes
        if not notes_password:
            # Save password
            GSettings.set_secret("errands_secret_notes", password)
            # Create notes table
            UserData.execute(
                """
                CREATE TABLE IF NOT EXISTS secret_notes (
                text TEXT,
                uid TEXT NOT NULL
                )
                """
            )
            # Open notes
            self.notes_win.set_content(SecretNotesPage(password))
            return

        # Verify password within 3 attempts and open notes
        if notes_password != password:
            # Delete data after 3 failed attempts
            if self.attempts > 2:
                UserData.execute("DROP TABLE secret_notes")
                GSettings.delete_secret("errands_secret_notes")
                self.delete_btn.set_visible(False)
                self.status_page.set_description(_("Create password"))
            # Update progressbar after failed attempt
            else:
                self.attempts += 1
                self.attempts_bar.set_fraction(self.attempts / 3)
                self.attempts_rev.set_reveal_child(True)
        else:
            # Open notes
            self.notes_win.set_content(SecretNotesPage(password))

    def _on_delete_clicked(self, btn: Gtk.Button):
        btn.set_visible(False)
        UserData.execute("DROP TABLE secret_notes")
        GSettings.delete_secret("errands_secret_notes")
        self.status_page.set_description(_("Create password"))


class SecretNotesWindow(Adw.Window):
    def __init__(self, window: Window):
        super().__init__()
        self.window: Window = window
        self._build_ui()

    def _build_ui(self):
        self.props.width_request = 360
        self.props.height_request = 360
        self.set_hide_on_close(True)
        self.set_title(_("Secret Notes"))
        GSettings.bind("secret-notes-width", self, "default_width")
        GSettings.bind("secret-notes-height", self, "default_height")

        self.set_content(SecretNotesPasswordPage(self))



# Copyright 2023 Vlad Krupinskii <mrvladus@yandex.ru>
# SPDX-License-Identifier: MIT

from gi.repository import GLib, Gio, Gtk, Secret  # type:ignore
from __main__ import APP_ID
from errands.lib.logging import Log

SECRETS_SCHEMA: Secret.Schema = Secret.Schema.new(
    APP_ID,
    Secret.SchemaFlags.NONE,
    {
        "account": Secret.SchemaAttributeType.STRING,
    },
)


class GSettings:
    """Class for accessing gsettings"""

    gsettings: Gio.Settings = None
    initialized: bool = False

    def _check_init(self):
        if not self.initialized:
            self.init()

    @classmethod
    def bind(
        self, setting: str, obj: Gtk.Widget, prop: str, invert: bool = False
    ) -> None:
        self._check_init(self)
        self.gsettings.bind(
            setting,
            obj,
            prop,
            (
                Gio.SettingsBindFlags.INVERT_BOOLEAN
                if invert
                else Gio.SettingsBindFlags.DEFAULT
            ),
        )

    @classmethod
    def get(self, setting: str):
        self._check_init(self)
        return self.gsettings.get_value(setting).unpack()

    @classmethod
    def set(self, setting: str, gvariant: str, value) -> None:
        self._check_init(self)
        self.gsettings.set_value(setting, GLib.Variant(gvariant, value))

    @classmethod
    def get_secret(self, account: str):
        self._check_init(self)
        return Secret.password_lookup_sync(SECRETS_SCHEMA, {"account": account}, None)

    @classmethod
    def set_secret(self, account: str, secret: str) -> None:
        self._check_init(self)

        return Secret.password_store_sync(
            SECRETS_SCHEMA,
            {
                "account": account,
            },
            Secret.COLLECTION_DEFAULT,
            f"Errands account credentials for {account}",
            secret,
            None,
        )

    @classmethod
    def delete_secret(self, account: str) -> bool:
        return Secret.password_clear_sync(SECRETS_SCHEMA, {"account": account}, None)

    @classmethod
    def init(self) -> None:
        Log.debug("Initialize GSettings")
        self.initialized = True
        self.gsettings = Gio.Settings.new(APP_ID)

        # Migrate old password
        try:
            account: int = self.gsettings.get_int("sync-provider")
            password: str = self.gsettings.get_string("sync-password")
            if 0 < account < 3 and password:
                account = "Nextcloud" if account == 1 else "CalDAV"
                self.set_secret(account, password)
                self.gsettings.set_string("sync-password", "")  # Clean pass
        except:
            pass
